"""
    Copyright 2022 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

"""
This module contains tests related to database concurrency issues. Whenever we fix a concurrency issue, be it performance
or deadlock related, a test should be added to ensure this occurence can not accidentally be introduced again.
"""

import asyncio
import datetime
import uuid
from collections import abc
from typing import Optional, Type, TypeVar

import asyncpg
import pytest

from inmanta import const, data
from inmanta.data import model
from inmanta.protocol.common import Result


def slowdown_queries(
    monkeypatch,
    *,
    cls: Type[data.BaseDocument] = data.BaseDocument,
    query_funcs: Optional[abc.Collection[str]] = None,
    delay: float = 1,
) -> None:
    """
    Introduces an artificial delay after each query execution through data.Document in order to increase the likelyhood of
    concurrent transactions.

    :param cls: The class to slow down the methods for.
    :param query_funcs: The names of the methods to add a delay to.
    :param delay: The amount of seconds to delay.
    """
    query_funcs = (
        query_funcs
        if query_funcs
        else [
            "select_query",
            "_fetchval",
            "_fetch_int",
            "_fetchrow",
            "_fetch_query",
            "_execute_query",
            "insert_many",
        ]
    )

    F = TypeVar("F", bound=abc.Coroutine)

    def patch_method(method: F) -> F:
        clsmethod: bool = hasattr(method, "__func__")
        func: F = method.__func__ if clsmethod else method

        async def patched(*args, **kwargs) -> object:
            # call unbound original method with cls object bound to the patched method
            result: object = await func(*args, **kwargs)
            await asyncio.sleep(delay)
            return result

        return classmethod(patched) if clsmethod else patched

    for query_func in query_funcs:
        monkeypatch.setattr(cls, query_func, patch_method(getattr(cls, query_func)))


@pytest.mark.slowtest
@pytest.mark.parametrize("endpoint_to_use", ["resource_deploy_done", "resource_action_update"])
async def test_4889_deadlock_delete_resource_action_update(
    monkeypatch, server, client, environment: str, agent, endpoint_to_use: str
) -> None:
    """
    Verify that no deadlock exists between the delete of a version and the deploy_done/resource_action_update (background task)
    on that same version.
    """
    env_id: uuid.UUID = uuid.UUID(environment)

    version: int = 1
    await data.ConfigurationModel(
        environment=env_id,
        version=version,
        date=datetime.datetime.now().astimezone(),
        total=1,
        version_info={},
        is_suitable_for_partial_compiles=False,
    ).insert()

    resource = model.ResourceVersionIdStr(f"std::File[agent1,path=/etc/file1],v={version}")
    await data.Resource.new(
        environment=env_id,
        status=const.ResourceState.available,
        resource_version_id=resource,
        attributes={"purge_on_delete": False, "purged": True, "requires": []},
    ).insert()

    # Add parameter for resource
    parameter_id = "test_param"
    result = await client.set_param(
        tid=env_id,
        id=parameter_id,
        source=const.ParameterSource.user,
        value="val",
        resource_id="std::File[agent1,path=/etc/file1]",
    )
    assert result.code == 200

    action_id = uuid.uuid4()
    result = await agent._client.resource_deploy_start(tid=env_id, rvid=resource, action_id=action_id)
    assert result.code == 200, result.result

    # artificially slow down queries to increase deadlock probability
    slowdown_queries(monkeypatch, cls=data.ResourceAction, query_funcs=["set_and_save"], delay=2)

    # request delete
    async def delete() -> Result:
        # Make sure insert starts first so it can acquire its first lock.
        await asyncio.sleep(1)
        return await client.delete_version(tid=environment, id=version)

    # request deploy_done
    now: datetime.datetime = datetime.datetime.now()
    deploy_done: abc.Awaitable[Result]
    if endpoint_to_use == "resource_deploy_done":
        deploy_done = agent._client.resource_deploy_done(
            tid=env_id,
            rvid=resource,
            action_id=action_id,
            status=const.ResourceState.deployed,
            messages=[
                model.LogLine(level=const.LogLevel.DEBUG, msg="message", kwargs={"keyword": 123, "none": None}, timestamp=now),
                model.LogLine(level=const.LogLevel.INFO, msg="test", kwargs={}, timestamp=now),
            ],
            changes={"attr1": model.AttributeStateChange(current=None, desired="test")},
            change=const.Change.purged,
        )
    elif endpoint_to_use == "resource_action_update":
        deploy_done = agent._client.resource_action_update(
            tid=env_id,
            resource_ids=[resource],
            action_id=action_id,
            action=const.ResourceAction.deploy,
            started=None,
            finished=now,
            status=const.ResourceState.deployed,
            messages=[
                data.LogLine.log(level=const.LogLevel.DEBUG, msg="message", timestamp=now, keyword=123, none=None),
                data.LogLine.log(level=const.LogLevel.INFO, msg="test", timestamp=now),
            ],
            changes={resource: {"attr1": model.AttributeStateChange(current=None, desired="test")}},
            change=const.Change.purged,
            send_events=True,
        )
    else:
        raise ValueError("Unknown value for endpoint_to_use parameter")

    # wait for both concurrent requests
    results: abc.Sequence[Result] = await asyncio.gather(deploy_done, delete())
    assert all(result.code == 200 for result in results), "\n".join(
        str(result.result) for result in results if result.code != 200
    )


@pytest.mark.slowtest
async def test_4889_deadlock_delete_resource_action_insert(monkeypatch, environment: str) -> None:
    """
    Verify that no deadlock exists between the delete of a version and the insert of a ResourceAction for that same version.
    """
    env_id: uuid.UUID = uuid.UUID(environment)

    version: int = 1
    confmodel: data.ConfigurationModel = data.ConfigurationModel(
        environment=env_id,
        version=version,
        date=datetime.datetime.now().astimezone(),
        total=1,
        version_info={},
        is_suitable_for_partial_compiles=False,
    )
    await confmodel.insert()

    resource = model.ResourceVersionIdStr(f"mymod::myresource[myagent,id=1],v={version}")
    await data.Resource.new(
        environment=env_id,
        status=const.ResourceState.available,
        resource_version_id=resource,
        attributes={},
    ).insert()

    # artificially slow down ResourceAction queries to increase deadlock probability
    slowdown_queries(monkeypatch, cls=data.ResourceAction, delay=1)

    insert: abc.Awaitable[None] = data.ResourceAction(
        environment=env_id,
        version=version,
        resource_version_ids=[resource],
        action_id=uuid.uuid4(),
        action=const.ResourceAction.deploy,
        started=datetime.datetime.now(),
    ).insert()

    async def delete() -> None:
        # Make sure insert starts first so it can acquire its first lock.
        await asyncio.sleep(0.5)
        await confmodel.delete_cascade()

    # Verify that this does not raise a deadlock exception. A failure on the insert is expected and acceptable if the delete
    # wins the race.
    try:
        await asyncio.gather(insert, delete())
    except asyncpg.ForeignKeyViolationError:
        pass



"""
    Copyright 2020 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from functools import reduce
from itertools import chain, filterfalse
from typing import TYPE_CHECKING, Callable, Dict, FrozenSet, Generic, Iterable, Iterator, List, Optional, Set, Tuple, TypeVar

if TYPE_CHECKING:
    from inmanta.ast import Locatable
    from inmanta.ast.attribute import Attribute
    from inmanta.ast.entity import Entity
    from inmanta.ast.statements.generator import Constructor
    from inmanta.execute.runtime import Resolver, ResultVariable


"""
This module contains a formal representation of the data flow modelled in the configuration model. The key elements of this
representation are detailed below.

DataflowGraph:
    The graph itself. An instance contains named nodes, which represent variables in the configuration model. DataflowGraph
    instances are nested. Each corresponds to a dynamic execution context (inmanta.execute.runtime.Resolver).

    The root DataflowGraph stores information about entity instances. It keeps track of all instance nodes for each entity and
    informs them of bidirectional relations and index matches. The graph needs to be informed in each of the following cases:
        - a new instance is created
        - a pair of bidirectional attributes is declared
        - an index match is detected

Node:
    A node in the graph represents a value, block variable or instance attribute in the configuration model. Nodes are
    connected by Assignment instances, the directed edges of the graph. There are five types of nodes:

    ValueNode:
        Represents a value.
    AssignableNode:
        Represents something that can be assigned to like a variable or an attribute. Contains assignments to values, instances
        and other assignables. An AssignableNode is part of an Equivalence, more on that below.
    AttributeNode:
        Special case of AssignableNode. Does some additional work on assignment to support bidirectional relations.
    InstanceNode:
        Represents an instance in the configuration model (or a tentative instance that hasn't been explicitly initialized yet,
        more on this in the Key methods section). Contains AttributeNodes.
        InstanceNode provides an index_match() method which registers another InstanceNode as an index match of the first one.
        From then on, all attribute-related calls to the first will be proxied to the second.
    NodeStub:
        Represents currently unsupported language elements so we can work around them without crashing.

NodeReference:
    A reference to a node. Since attributes are part of an instance but we usually refer to them indirectly on a variable, a
    level of indirection is required in the graph as well. For example, consider the following statements:
        n = x.n
        x = A(n = 42)
    In this case x.n refers to the attribute n of the instance of A.

    Types of NodeReference:

    ValueNodeReference:
        Refers to a ValueNode directly.
    AssignableNodeReference:
        Represents NodeReferences that can be assigned to. Parent for VariableNodeReference and AttributeNodeReference.
    VariableNodeReference:
        Refers to an AssignableNode directly.
    AttributeNodeReference:
        Refers to an AssignableNode indirectly by specifying an attribute on another AssignableNodeReference i. Points to an
        AttributeNode of an InstanceNode on the end of the assignment chain starting in i.
    InstanceNodeReference:
        Refers to an InstanceNode directly.

Assignment:
    Represents assignment, possibly indirect, in the configuration model. Constructors for example are also modelled as
    assignment of the individual attributes to their values. An Assignment instance consists of a lhs (AssignableNode) and a
    rhs (NodeReference), a responsible (the Locatable responsible for the creation of this edge) and a context (the
    DataflowGraph this assignment was created in).

    A note on indirection in the rhs and direction in the lhs:
        The primary goal of the dataflow graph is to be able to reason about the values of a variable and how it received those
        values. Therefore it makes sense to centralize outgoing value assignments (the lhs) directly on the node, while keeping
        indirection in the rhs. For example, in the model
            x = y
            x.n = u.n
            y.n = 42
        the assignment to x.n would be modelled by resolving the node reference x.n to a specific node (a more detailed
        explanation about this resolution can be found in the Key methods section). The right hand side u.n however would be
        modelled by keeping the indirection and just storing the node reference u.n. By resolving the left hand side to a node
        we make sure that the assignment to `x.n` and that to `y.n`, which are two references to the same variable, are modelled
        in the same location, namely the attribute node on y.

Equivalence:
    The Equivalence class is introduced to efficiently handle assignment loops. Consider the model
        x = y
        y = z
        z = x
        x = 42
    This is a valid configuration model where each of the variables receives the value 42. An equivalence is defined as a
    collection of nodes that are fully equivalent to another as a result of one or more assignment loops. This means that
    assignment to one of them, has effect on all of them. Because leaf finding is such an important part of the modelling
    process it makes sense to make abstraction of loop detection. Loop detection is only performed when a new edge is
    introduced, based on existing equivalences. If a loop is found, it is cached as the Equivalence of each of it's nodes.

    Apart from that, an Equivalence is where tentative instances are stored. Tentative instances are explained in the Key
    methods section.

    Invariants:
        n.equivalence == e <=> n in e.nodes
        n.equivalence == e <=> n' in e.nodes for n' in loop_detect(n)

Key methods:
    leaves() in AssignableNodeReference, AssignableNode, Equivalence:
        The DataflowGraph is a directed graph where edges represent assignment. As a result a variable's value, or cause for the
        lack thereof, can be found by following the edges until a value assignment is reached, or until a node is reached that
        has no outgoing edges. The term used to refer to these nodes (with either at least one outgoing edge to a value or
        instance node or no outgoing edges at all) throughout this module is leaf.
        In practice, it's slightly more complicated. Because of indirection in the edges (Assignment's rhs is a NodeReference,
        not a Node), at some point in time a NodeReference might exist that can not be reduced to any Node. For example `x.n` if
        no instance has been assigned to x. In that case the reference itself is a leaf.

    AssignableNodeReference.assignment_node():
        Because of indirection, an AssignableNodeReference might point to multiple AssignableNodes. For a simple VariableNode
        this is never the case, but for a AttributeNodeReference this might be relevant. Consider the following statements:
            x = y
            x = z
        In this case it's not clear what node x.n would refer to. So how should we handle an assignment `x.n = 42`? That's
        exactly what this method is for: it returns the AssignableNode to which an assignment to the reference should be
        forwarded. An important consideration here is that if the assignments branch out, they will either come back together
        later (for example because of `y = z` or an index match), or a double assignment error will occur on x. Which node
        should we model the assignment `x.n = 42` on in both cases?
            Assignment tree comes back together:
                We can pick any node that is currently a leaf of the tree originating from x.n. Since they come back together
                it has no effect which node we pick.
            Double assignment error occurs on x:
                There is a double assignment on x so the exact behavior of x.n is not that pressing. With that in mind, there is
                no reason to prefer one leaf node over another.
        Conclusion: we can pick any leaf node to model the assignment on. In case of divergence, either it will fix itself by
        converging again, or there are more pressing problems in the model.
        If no leaf node currently exists, we create a tentative InstanceNode on the leaf node of the tree originating in x
        (not x.n) and return the attribute n of that InstanceNode as the assignment node. On later assignments to x the
        tentative instance will get propagated so that it always exists on a leaf until it finally merges with an actual
        instance, instantiated somewhere in the assignment tree originating from x. If it doesn't merge that means x is not an
        instance and a compilation error will occur.
"""


class DataflowGraph:
    """
    Graph representation of the data flow and declarative control flow of an Inmanta model.
    """

    __slots__ = ("resolver", "parent", "_own_variables", "_own_instances")

    def __init__(self, resolver: "Resolver", parent: Optional["DataflowGraph"] = None) -> None:
        self.resolver: "Resolver" = resolver
        self.parent: Optional[DataflowGraph] = parent if parent is not None else None
        # keeps track of variables that have not been declared in the resolver's scope. This should only be populated
        # if the model refers to a variable in the rhs that has no declaration in the left hand side.
        # For example `n = y.n` in a scope that does not contain a `y = ...` statement.
        self._own_variables: Dict[str, AssignableNode] = {}
        # keeps track of instance nodes and their responsible
        self._own_instances: Dict["Constructor", InstanceNode] = {}

    def get_own_variable(self, name: str) -> "AssignableNodeReference":
        """
        Returns a reference to one of this graph's own variable nodes.
        """
        if name not in self._own_variables:
            self._own_variables[name] = AssignableNode(name)
        return self._own_variables[name].reference()

    def own_instance_node_for_responsible(
        self, entity: "Entity", responsible: "Constructor", get_new: Callable[[], "InstanceNode"]
    ) -> "InstanceNode":
        """
        Returns this graph's instance node tied to responsible if it exists.
        Otherwise, creates a new one using get_new, returns it and adds it to the graph.
        """
        if responsible not in self._own_instances:
            new: InstanceNode = get_new()
            new.entity = entity
            new.responsible = responsible
            new.context = self
            self._own_instances[responsible] = new
        return self._own_instances[responsible]

    def add_index_match(self, instances: Iterable["InstanceNodeReference"]) -> None:
        """
        Registers an index match between a set of instance nodes. Informs the nodes of the match.
        """
        iterator: Iterator[InstanceNodeReference] = iter(instances)
        try:
            first: InstanceNode = next(iter(iterator)).node()
            for instance in iterator:
                instance.node().index_match(first)
        except StopIteration:
            pass


class Node:
    """
    A node in the data flow graph. Represents an attribute, variable, value or instance in the configuration model.
    """

    __slots__ = ()

    def __init__(self) -> None:
        pass

    def reference(self) -> "NodeReference":
        """
        Returns a reference to this node.
        """
        raise NotImplementedError()


class NodeReference:
    """
    Reference to a node.
    """

    __slots__ = ()

    def __init__(self) -> None:
        pass

    def nodes(self) -> Iterator["Node"]:
        """
        Returns all nodes this NodeReference refers to.
        """
        raise NotImplementedError()

    def ref_to_node(self, node: "Node") -> bool:
        """
        Returns true iff this NodeReference refers to node.
        """
        return node in self.nodes()

    def assign_to(self, lhs: "AssignableNode", responsible: "Locatable", context: "DataflowGraph") -> None:
        """
        Assigns this node to the left hand side node.
        """
        raise NotImplementedError()

    def assign_attribute(
        self, attribute: str, rhs: "NodeReference", responsible: "Locatable", context: "DataflowGraph"
    ) -> None:
        raise NotImplementedError()


class AssignableNodeReference(NodeReference):
    """
    Reference to a node that can have values assigned to it.
    """

    __slots__ = ()

    def __init__(self) -> None:
        NodeReference.__init__(self)

    def nodes(self) -> Iterator["AssignableNode"]:
        raise NotImplementedError()

    def leaves(self) -> Iterator["AssignableNodeReference"]:
        """
        Returns an iterator over this reference's leaves. A leaf is defined as a node reference
        in the assignment tree which has either one or more value or instance assignments
        or no assignments at all. This reference's leaves are the leaves of the assignment
        subtrees originating from this reference's nodes.
        """
        child_leaves: Iterator["AssignableNodeReference"] = chain.from_iterable(n.leaves() for n in self.nodes())
        try:
            yield next(child_leaves)
            yield from child_leaves
        except StopIteration:
            yield self

    def leaf_nodes(self) -> Iterator["AssignableNode"]:
        """
        Returns an iterator over the nodes this reference's leaves refer to.
        """
        return chain.from_iterable(leaf.nodes() for leaf in self.leaves())

    def any_leaf_node(self) -> Optional["AssignableNode"]:
        """
        Returns a single leaf for this reference, if any exist.
        """
        try:
            return next(self.leaf_nodes())
        except StopIteration:
            return None

    def assignment_node(self) -> "AssignableNode":
        """
        Returns the assignable node to which an assignment to this reference should be delegated.
        """
        raise NotImplementedError()

    def assign(self, node_ref: "NodeReference", responsible: "Locatable", context: "DataflowGraph") -> None:
        """
        Assign another node to this reference's assignment node.
        """
        self.assignment_node().assign(node_ref, responsible, context)

    def assign_to(self, lhs: "AssignableNode", responsible: "Locatable", context: "DataflowGraph") -> None:
        lhs.assign_assignable(self, responsible, context)

    def get_attribute(self, name: str) -> "AttributeNodeReference":
        """
        Returns a reference to an attribute of this reference's node, by name.
        """
        return AttributeNodeReference(self, name)

    def assign_attribute(
        self, attribute: str, rhs: "NodeReference", responsible: "Locatable", context: "DataflowGraph"
    ) -> None:
        self.get_attribute(attribute).assign(rhs, responsible, context)

    def set_result_variable(self, result_variable: "ResultVariable") -> None:
        """
        Sets this nodes' result variable. It is sufficient to set it once
        because at the point a ResultVariable gets created, the corresponding
        node exists and should not change except when explicitly replaced by this
        module. In that case it's the responsibility of the actor to propagate
        the result variable.
        """
        for node in self.nodes():
            node.set_result_variable(result_variable)


class AttributeNodeReference(AssignableNodeReference):
    """
    Reference to a node representing an attribute of another node.
    """

    __slots__ = ("instance_var_ref", "attribute")

    def __init__(self, instance_var_ref: AssignableNodeReference, attribute: str) -> None:
        AssignableNodeReference.__init__(self)
        self.instance_var_ref: AssignableNodeReference = instance_var_ref
        self.attribute: str = attribute

    def nodes(self) -> Iterator["AttributeNode"]:
        # yield all attribute nodes on instances assigned to this reference's leaves
        for node in self.instance_var_ref.leaf_nodes():
            for instance_node in chain(
                (assignment.rhs.node() for assignment in node.instance_assignments),
                [node.equivalence.tentative_instance] if node.equivalence.tentative_instance is not None else [],
            ):
                yield instance_node.register_attribute(self.attribute)

    def assignment_node(self) -> "AssignableNode":
        try:
            return next(self.nodes())
        except StopIteration:
            # If no attribute node can be found, create a tentative one on a leaf of the instance_var_ref.
            # Tentative attributes of a VariableNode will be propagated on assignment to that node.
            instance_leaf: Optional["AssignableNode"] = self.instance_var_ref.any_leaf_node()
            if instance_leaf is None:
                instance_leaf = self.instance_var_ref.assignment_node()
            return instance_leaf.equivalence.tentative_attribute(self.attribute)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AttributeNodeReference):
            return NotImplemented
        return self.instance_var_ref == other.instance_var_ref and self.attribute == other.attribute

    def __repr__(self) -> str:
        return "%s.%s" % (repr(self.instance_var_ref), self.attribute)


class InstanceAttributeNodeReference(AssignableNodeReference):
    """
    Reference to a node representing an attribute of an instance.
    """

    __slots__ = ("instance", "attribute")

    def __init__(self, instance: "InstanceNode", attribute: str) -> None:
        AssignableNodeReference.__init__(self)
        self.instance: InstanceNode = instance
        self.attribute: str = attribute

    def nodes(self) -> Iterator["AttributeNode"]:
        yield self.instance.register_attribute(self.attribute)

    def assignment_node(self) -> "AttributeNode":
        return next(self.nodes())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, InstanceAttributeNodeReference):
            return NotImplemented
        return self.assignment_node() == other.assignment_node()

    def __repr__(self) -> str:
        return repr(self.assignment_node())


class DirectNodeReference(NodeReference):
    """
    Direct reference to a Node.
    """

    __slots__ = "node"

    def __init__(self, node: "Node") -> None:
        NodeReference.__init__(self)
        self.node: Node = node

    def nodes(self) -> Iterator["Node"]:
        yield self.node

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DirectNodeReference):
            return NotImplemented
        return self.node == other.node


class VariableNodeReference(AssignableNodeReference, DirectNodeReference):
    """
    Reference to a node representing a variable in the configuration model.
    """

    __slots__ = ()

    def __init__(self, node: "AssignableNode") -> None:
        AssignableNodeReference.__init__(self)
        DirectNodeReference.__init__(self, node)
        self.node: AssignableNode

    def nodes(self) -> Iterator["AssignableNode"]:
        yield self.node

    def assignment_node(self) -> "AssignableNode":
        return self.node

    def __repr__(self) -> str:
        return repr(self.node)


class ValueNodeReference(DirectNodeReference):
    """
    Reference to a node representing a value in the configuration model.
    """

    __slots__ = ()

    def __init__(self, node: "ValueNode") -> None:
        DirectNodeReference.__init__(self, node)
        self.node: ValueNode

    def assign_to(self, lhs: "AssignableNode", responsible: "Locatable", context: "DataflowGraph") -> None:
        lhs.assign_value(self, responsible, context)

    def assign_attribute(
        self, attribute: str, rhs: "NodeReference", responsible: "Locatable", context: "DataflowGraph"
    ) -> None:
        if isinstance(self.node, NodeStub):
            return
        raise Exception("Can not assign attribute on a value node")

    def __repr__(self) -> str:
        return repr(self.node)


class InstanceNodeReference(NodeReference):
    """
    Reference to a node representing an entity instance in the configuration model.
    """

    __slots__ = "_node"

    def __init__(self, node: "InstanceNode") -> None:
        NodeReference.__init__(self)
        self._node: InstanceNode = node

    def top_node(self) -> "InstanceNode":
        """
        Returns the node this reference refers to directly.
        """
        return self._node

    def node(self) -> "InstanceNode":
        """
        Returns the actual instance node in case of index matches.
        """
        return self.top_node().get_self()

    def nodes(self) -> Iterator["Node"]:
        yield self.node()

    def assign_attribute(
        self, attribute: str, rhs: "NodeReference", responsible: "Locatable", context: "DataflowGraph"
    ) -> None:
        """
        Assigns a node to an attribute of the instance this reference refers to.
        """
        self.node().assign_attribute(attribute, rhs, responsible, context)

    def assign_to(self, lhs: "AssignableNode", responsible: "Locatable", context: "DataflowGraph") -> None:
        lhs.assign_instance(self, responsible, context)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, InstanceNodeReference):
            return NotImplemented
        return self._node == other._node

    def __repr__(self) -> str:
        return repr(self.top_node())


RT = TypeVar("RT", bound=NodeReference, covariant=True)


class Assignment(Generic[RT]):
    """
    Assignment of one node to another, caused by a responsible

    :param lhs: The left hand side of the assignment, a node.
    :param rhs: The right hand side of the assignment, a reference to a node.
    :param responsible: The responsible (statement) for this assignment.
    :param context: The dynamic context (e.g. implementation) this assignment lives in.
    """

    __slots__ = ("lhs", "rhs", "responsible", "context")

    def __init__(self, lhs: "AssignableNode", rhs: RT, responsible: "Locatable", context: "DataflowGraph") -> None:
        self.lhs: "AssignableNode" = lhs
        self.rhs: RT = rhs
        self.responsible: "Locatable" = responsible
        self.context: "DataflowGraph" = context


class ValueNode(Node):
    """
    Node representing a value in the configuration model.
    """

    __slots__ = "value"

    def __init__(self, value: object) -> None:
        super().__init__()
        self.value: object = value

    def reference(self) -> "ValueNodeReference":
        return ValueNodeReference(self)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ValueNode):
            return NotImplemented
        return self.value == other.value

    def __hash__(self) -> int:
        return hash(self.value)

    def __repr__(self) -> str:
        return repr(self.value)


class NodeStub(ValueNode):
    """
    Node that represents currently unsupported expressions.
    """

    __slots__ = "message"

    def __init__(self, message: str) -> None:
        ValueNode.__init__(self, message)
        self.message = message

    def reference(self) -> "ValueNodeReference":
        return ValueNodeReference(self)


class AssignableNode(Node):
    """
    Node representing a variable or an attribute in the assignment graph model.
    """

    __slots__ = (
        "name",
        "assignable_assignments",
        "value_assignments",
        "instance_assignments",
        "equivalence",
        "result_variable",
    )

    def __init__(self, name: str) -> None:
        Node.__init__(self)
        self.name: str = name
        self.assignable_assignments: List[Assignment[AssignableNodeReference]] = []
        self.value_assignments: List[Assignment[ValueNodeReference]] = []
        self.instance_assignments: List[Assignment[InstanceNodeReference]] = []
        self.equivalence: Equivalence = Equivalence(frozenset([self]))
        self.result_variable: Optional[ResultVariable] = None

    def reference(self) -> AssignableNodeReference:
        return VariableNodeReference(self)

    def leaves(self) -> Iterator["AssignableNodeReference"]:
        """
        Returns an iterator over this node's leaves. A leaf is defined as a node
        in the assignment tree which has either one or more value or instance assignments
        or no assignments at all. This node's leaves are the leaves of the assignment
        subtree originating from this node.
        """
        return self.equivalence.leaves()

    def assignments(self) -> Iterator[Assignment]:
        """
        Returns an iterator over assignments to this node.
        """
        return chain(self.assignable_assignments, self.value_assignments, self.instance_assignments)

    def assign(self, node_ref: "NodeReference", responsible: "Locatable", context: "DataflowGraph") -> None:
        """
        Assigns another node to this one, by reference.
        """
        node_ref.assign_to(self, responsible, context)

    def assign_value(self, val_ref: ValueNodeReference, responsible: "Locatable", context: "DataflowGraph") -> None:
        """
        Assigns a value node to this node, by reference.
        """
        self.value_assignments.append(Assignment(self, val_ref, responsible, context))

    def assign_instance(self, instance_ref: InstanceNodeReference, responsible: "Locatable", context: "DataflowGraph") -> None:
        """
        Assigns an instance node to this node, by reference.
        """
        self.instance_assignments.append(Assignment(self, instance_ref, responsible, context))
        self.equivalence.propagate_tentative_instance()

    def assign_assignable(self, var_ref: AssignableNodeReference, responsible: "Locatable", context: "DataflowGraph") -> None:
        """
        Assigns an assignable node to this node, by reference.
        """
        # Gather all equivalences on the path between the rhs's leaves and this node.
        # The existence of such a trail indicates an assignment loop.
        equivalence_trail: Set[Equivalence] = reduce(
            set.union,
            (n.equivalence.equivalences_on_path(self) for n in var_ref.nodes()),
            set(),
        )
        # merge all equivalences on the trail
        new_equivalence: Equivalence = reduce(
            lambda acc, eq: acc.merge(eq),
            equivalence_trail,
            Equivalence(frozenset()),
        )
        self.assignable_assignments.append(Assignment(self, var_ref, responsible, context))
        # apply new equivalence to all it's nodes
        for node in new_equivalence.nodes:
            node.equivalence = new_equivalence
        # propagate this node's tentative instance to the new leaves, if it exists
        self.equivalence.propagate_tentative_instance()

    def set_result_variable(self, result_variable: "ResultVariable") -> None:
        assert self.result_variable is None or self.result_variable is result_variable
        self.result_variable = result_variable

    def __repr__(self) -> str:
        return self.name


class Equivalence:
    """
    Represents a collection of nodes that are equivalent because of one or more assignment loops.
    """

    __slots__ = ("nodes", "tentative_instance")

    def __init__(
        self, nodes: FrozenSet[AssignableNode] = frozenset(), tentative_instance: Optional["InstanceNode"] = None
    ) -> None:
        self.nodes: FrozenSet[AssignableNode] = nodes
        self.tentative_instance: Optional[InstanceNode] = tentative_instance

    def merge(self, other: "Equivalence") -> "Equivalence":
        """
        Returns the equivalence that is the union of this one and the one passed as an argument.
        """
        tentative_instance: Optional[InstanceNode] = self.tentative_instance
        if tentative_instance is not None:
            self.tentative_instance = None
            if other.tentative_instance is not None:
                tentative_instance.merge(other.tentative_instance)
        return Equivalence(self.nodes.union(other.nodes), tentative_instance)

    def is_leaf(self) -> bool:
        """
        Returns true iff this equivalence is a leaf. An equivalence is a leaf iff it has an instance
        or value assignment, or doesn't have any assignments at all.
        """
        return (
            any(self.instance_assignments()) or any(self.value_assignments()) or not any(self.external_assignable_assignments())
        )

    def leaves(self) -> Iterator[AssignableNodeReference]:
        """
        Returns an iterator over all leaves on assignment paths originating from this equivalence.
        """
        if self.is_leaf():
            explicit_leaves: Iterator[AssignableNodeReference] = (
                node.reference() for node in self.nodes if len(node.value_assignments) > 0 or len(node.instance_assignments) > 0
            )
            try:
                yield next(explicit_leaves)
                yield from explicit_leaves
            except StopIteration:
                yield from (node.reference() for node in self.nodes)
        yield from (node for assignment in self.external_assignable_assignments() for node in assignment.rhs.leaves())

    def equivalences_on_path(self, node: AssignableNode) -> Set["Equivalence"]:
        """
        Returns the set of all equivalences on the assignment path originating from this equivalence
        and terminating in the equivalence containing node.
        Returns the empty set if there is no such path.
        """
        if node in self.nodes:
            return {self}
        child_paths: Set[Equivalence] = reduce(
            set.union,
            (
                n.equivalence.equivalences_on_path(node)
                for n in set(chain.from_iterable(a.rhs.nodes() for a in self.external_assignable_assignments()))
            ),
            set(),
        )
        return child_paths.union({self}) if len(child_paths) > 0 else set(())

    def _is_internal_assignment(self, assignment: Assignment[AssignableNodeReference]) -> bool:
        filtered: Iterator[AssignableNode] = filter(assignment.rhs.ref_to_node, self.nodes)
        return any(filtered)

    def interal_assignments(self) -> Iterator[Assignment[AssignableNodeReference]]:
        """
        Returns an iterator over internal (rhs is part of this equivalence) AssignableNodes assigned to this Equivalence.
        """
        return filter(self._is_internal_assignment, chain.from_iterable(n.assignable_assignments for n in self.nodes))

    def external_assignable_assignments(self) -> Iterator[Assignment[AssignableNodeReference]]:
        """
        Returns an iterator over external (rhs is not part of this equivalence) AssignableNodes assigned to this
        Equivalence.
        """
        return filterfalse(self._is_internal_assignment, chain.from_iterable(n.assignable_assignments for n in self.nodes))

    def instance_assignments(self) -> Iterator[Assignment[InstanceNodeReference]]:
        """
        Returns an iterator over InstanceNodes assigned to this Equivalence.
        """
        return chain.from_iterable(n.instance_assignments for n in self.nodes)

    def value_assignments(self) -> Iterator[Assignment[ValueNodeReference]]:
        """
        Returns an iterator over ValueNodes assigned to this Equivalence.
        """
        return chain.from_iterable(n.value_assignments for n in self.nodes)

    def propagate_tentative_instance(self) -> None:
        """
        Propagates this equivalence's tentative instance to one of it's leaves.
        """

        def propagate_tentative_assignments(source: Equivalence, target: AssignableNodeReference) -> None:
            """
            Recursively propagate tentative assignments.
            Makes sure tentative assignments over a relation get propageted as well.
            """
            if source.tentative_instance is None:
                return
            for tent_attr_name, tent_attr_node in source.tentative_instance.attributes.items():
                attr_node: AttributeNodeReference = target.get_attribute(tent_attr_name)
                for assignment in tent_attr_node.assignments():
                    attr_node.assign(assignment.rhs, assignment.responsible, assignment.context)
                propagate_tentative_assignments(tent_attr_node.equivalence, target.get_attribute(tent_attr_name))
            self.tentative_instance = None

        try:
            leaf: AssignableNodeReference = next(self.leaves())
            propagate_tentative_assignments(self, leaf)
        except StopIteration:
            raise Exception("Inconsistent state: an equivalence should always have at least one leaf")

    def tentative_attribute(self, attr: str) -> "AttributeNode":
        if self.tentative_instance is None:
            self.tentative_instance = InstanceNode([])
        attr_node: Optional[AttributeNode] = self.tentative_instance.get_attribute(attr)
        if attr_node is None:
            attr_node = self.tentative_instance.register_attribute(attr)
        return attr_node

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Equivalence):
            return NotImplemented
        return self.nodes == other.nodes and self.tentative_instance == other.tentative_instance

    def __hash__(self) -> int:
        return hash(self.nodes)


class AttributeNode(AssignableNode):
    """
    Node representing an entity's attribute.
    """

    __slots__ = ("instance", "responsibles")

    def __init__(self, instance: "InstanceNode", name: str) -> None:
        AssignableNode.__init__(self, name)
        self.instance: InstanceNode = instance
        self.responsibles: Set[Tuple["Locatable", DataflowGraph]] = set(())

    def assign(self, node_ref: NodeReference, responsible: "Locatable", context: "DataflowGraph") -> None:
        # only add assignment for each responsible once, this check is necessary for bidirectional attributes
        if (responsible, context) in self.responsibles:
            return
        super().assign(node_ref, responsible, context)
        self.responsibles.add((responsible, context))
        self.instance.assign_other_direction(self.name, node_ref, responsible, context)

    def __repr__(self) -> str:
        return "attribute %s on %s" % (super().__repr__(), repr(self.instance))


class InstanceNode(Node):
    """
    Node representing an entity instance.
    """

    __slots__ = (
        "attributes",
        "entity",
        "responsible",
        "context",
        "bidirectional_attributes",
        "_index_node",
        "_all_index_nodes",
    )

    def __init__(
        self,
        attributes: Iterable[str],
    ) -> None:
        Node.__init__(self)
        self.attributes: Dict[str, AttributeNode] = {name: AttributeNode(self, name) for name in attributes}
        self.entity: Optional["Entity"] = None
        self.responsible: Optional["Constructor"] = None
        self.context: Optional["DataflowGraph"] = None
        self._index_node: Optional[InstanceNode] = None
        self._all_index_nodes: Set["InstanceNode"] = {self}

    def reference(self) -> InstanceNodeReference:
        return InstanceNodeReference(self)

    def get_self(self) -> "InstanceNode":
        """
        Returns the main instance node if this node acts as a proxy due to an index match.
        Otherwise returns itself.
        """
        return self if self._index_node is None else self._index_node.get_self()

    def merge(self, other: "InstanceNode") -> None:
        """
        Merge another instance into this one.
        """
        if self.get_self() is not self:
            return self.get_self().merge(other)
        for attr_name, attr_node in other.get_self().attributes.items():
            for assignment in attr_node.assignments():
                self.assign_attribute(attr_name, assignment.rhs, assignment.responsible, assignment.context)
            if attr_node.result_variable is not None:
                self.register_attribute(attr_name).set_result_variable(attr_node.result_variable)

    def index_match(self, index_node: "InstanceNode") -> None:
        """
        Registers index_node as this node's index node. Propagates all attribute assignments.
        """
        if index_node is self:
            return
        if self._index_node is not None:
            raise Exception("Trying to match index on node that already has an index match. Try calling get_self() first")
        assert self.entity == index_node.get_self().entity
        index_node.get_self().merge(self)
        self._index_node = index_node
        self.attributes = {}
        index_node.get_self().update_all_index_nodes(self._all_index_nodes)
        self._all_index_nodes = set(())

    def update_all_index_nodes(self, index_nodes: Set["InstanceNode"]) -> None:
        """
        Adds a set of index nodes to this node's set of all index matches.
        """
        if self.get_self() is not self:
            return self.get_self().update_all_index_nodes(index_nodes)
        self._all_index_nodes.update(index_nodes)

    def get_all_index_nodes(self) -> Set["InstanceNode"]:
        """
        Returns all index matches for this node.
        """
        if self.get_self() is not self:
            return self.get_self().get_all_index_nodes()
        return self._all_index_nodes

    def assign_attribute(
        self,
        attribute: str,
        node_ref: "NodeReference",
        responsible: "Locatable",
        context: "DataflowGraph",
    ) -> None:
        """
        Assigns a node to one of this instance's attributes.
        """
        if self.get_self() is not self:
            return self.get_self().assign_attribute(attribute, node_ref, responsible, context)
        attr_node: AttributeNode = self.register_attribute(attribute)
        attr_node.assign(node_ref, responsible, context)

    def register_attribute(self, attribute: str) -> AttributeNode:
        """
        Registers an attribute to this instance node. Returns the attribute's node.
        """
        if self.get_self() is not self:
            return self.get_self().register_attribute(attribute)
        if attribute not in self.attributes:
            self.attributes[attribute] = AttributeNode(self, attribute)
        return self.attributes[attribute]

    def get_attribute(self, attribute: str) -> Optional[AttributeNode]:
        """
        Returns one of this instance's attributes by name, if it exists.
        """
        if self.get_self() is not self:
            return self.get_self().get_attribute(attribute)
        try:
            return self.attributes[attribute]
        except KeyError:
            return None

    def get_index_attributes(self) -> Iterator[AttributeNode]:
        if self.get_self() is not self:
            return self.get_self().get_index_attributes()
        assert self.entity is not None
        yield from (self.register_attribute(i) for i in chain.from_iterable(self.entity.get_indices()))

    def assign_other_direction(
        self, attribute: str, node_ref: "NodeReference", responsible: "Locatable", context: "DataflowGraph"
    ) -> None:
        """
        If attribute is a bidirectional attribute, assign the other direction.
        :param attribute: this node's attribute
        :param node_ref: the node assigned to the attribute
        :param responsible: the responsible for both assignments
        :param context: the context for both assignments
        """
        if self.get_self() is not self:
            return self.get_self().assign_other_direction(attribute, node_ref, responsible, context)
        if self.entity is None:
            return
        ast_attribute: Optional["Attribute"] = self.entity.get_attribute(attribute)
        if ast_attribute is None:
            return
        if ast_attribute.end is not None:
            node_ref.assign_attribute(ast_attribute.end.get_name(), self.reference(), responsible, context)

    def __repr__(self) -> str:
        return "%s instance" % self.entity



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import base64
import inspect
import logging
import traceback
import typing
import uuid
from abc import ABC, abstractmethod
from collections import abc, defaultdict
from concurrent.futures import Future
from typing import Any, Callable, Dict, Generic, List, Optional, Tuple, Type, TypeVar, Union, cast, overload

from tornado import concurrent

import inmanta
from inmanta import const, data, protocol, resources
from inmanta.agent import io
from inmanta.agent.cache import AgentCache
from inmanta.const import ParameterSource, ResourceState
from inmanta.data.model import AttributeStateChange, ResourceIdStr
from inmanta.protocol import Result, json_encode
from inmanta.stable_api import stable_api
from inmanta.types import SimpleTypes
from inmanta.util import hash_file

if typing.TYPE_CHECKING:
    import inmanta.agent.agent
    from inmanta.agent.io.local import IOBase


LOGGER = logging.getLogger(__name__)

T = TypeVar("T")
T_FUNC = TypeVar("T_FUNC", bound=Callable[..., Any])


@stable_api
class provider(object):  # noqa: N801
    """
    A decorator that registers a new handler.

    :param resource_type: The type of the resource this handler provides an implementation for.
                          For example, :inmanta:entity:`std::File`
    :param name: A name to reference this provider.
    """

    def __init__(self, resource_type: str, name: str) -> None:
        self._resource_type = resource_type
        self._name = name

    def __call__(self, function):
        """
        The wrapping
        """
        Commander.add_provider(self._resource_type, self._name, function)
        return function


@stable_api
class SkipResource(Exception):
    """
    A handler should raise this exception when a resource should be skipped. The resource will be marked as skipped
    instead of failed.
    """


@stable_api
class ResourcePurged(Exception):
    """
    If the :func:`~inmanta.agent.handler.CRUDHandler.read_resource` method raises this exception, the agent will
    mark the current state of the resource as purged.
    """


class InvalidOperation(Exception):
    """
    This exception is raised by the context or handler methods when an invalid operation is performed.
    """


@stable_api
def cache(
    func: Optional[T_FUNC] = None,
    ignore: typing.List[str] = [],
    timeout: int = 5000,
    for_version: bool = True,
    cache_none: bool = True,
    # deprecated parameter kept for backwards compatibility: if set, overrides cache_none
    cacheNone: Optional[bool] = None,  # noqa: N803
    call_on_delete: Optional[Callable[[Any], None]] = None,
) -> Union[T_FUNC, Callable[[T_FUNC], T_FUNC]]:
    """
    decorator for methods in resource handlers to provide caching

    this decorator works similar to memoization:
    when the decorate method is called, its return value is cached,
    for subsequent calls, the cached value is used instead of the actual value

    The name of the method + the arguments of the method form the cache key

    If an argument named version is present and for_version is True,
    the cache entry is flushed after this version has been deployed
    If an argument named resource is present,
    it is assumed to be a resource and its ID is used, without the version information

    :param timeout: the number of second this cache entry should live
    :param for_version: if true, this value is evicted from the cache when this deploy is ready
    :param ignore: a list of argument names that should not be part of the cache key
    :param cache_none: cache returned none values
    :param call_on_delete: A callback function that is called when the value is removed from the cache,
            with the value as argument.
    """

    def actual(f: Callable) -> T_FUNC:
        myignore = set(ignore)
        sig = inspect.signature(f)
        myargs = list(sig.parameters.keys())[1:]

        def wrapper(self, *args: object, **kwds: object) -> object:
            kwds.update(dict(zip(myargs, args)))

            def bound(**kwds):
                return f(self, **kwds)

            return self.cache.get_or_else(
                f.__name__,
                bound,
                for_version,
                timeout,
                myignore,
                cacheNone if cacheNone is not None else cache_none,
                **kwds,
                call_on_delete=call_on_delete,
            )

        # Too much magic to type statically
        return cast(T_FUNC, wrapper)

    if func is None:
        return actual
    else:
        return actual(func)


@stable_api
class LoggerABC(ABC):
    """
    Minimal logging interface exposing logging methods for commonly used
    logging levels.

    Child classes are responsible for implementing a _log_msg method with the
    concrete logging implementation.
    """

    def critical(self, msg: str, *args: object, exc_info: bool = False, **kwargs: object) -> None:
        self._log_msg(logging.CRITICAL, msg, *args, exc_info=exc_info, **kwargs)

    def error(self, msg: str, *args: object, exc_info: bool = False, **kwargs: object) -> None:
        self._log_msg(logging.ERROR, msg, *args, exc_info=exc_info, **kwargs)

    def warning(self, msg: str, *args: object, exc_info: bool = False, **kwargs: object) -> None:
        self._log_msg(logging.WARNING, msg, *args, exc_info=exc_info, **kwargs)

    def info(self, msg: str, *args: object, exc_info: bool = False, **kwargs: object) -> None:
        self._log_msg(logging.INFO, msg, *args, exc_info=exc_info, **kwargs)

    def debug(self, msg: str, *args: object, exc_info: bool = False, **kwargs: object) -> None:
        self._log_msg(logging.DEBUG, msg, *args, exc_info=exc_info, **kwargs)

    def exception(self, msg: str, *args: object, exc_info: bool = True, **kwargs: object) -> None:
        self.error(msg, *args, exc_info=exc_info, **kwargs)

    @abstractmethod
    def _log_msg(
        self,
        level: int,
        msg: str,
        *args: object,
        exc_info: bool = False,
        **kwargs: object,
    ) -> None:
        raise NotImplementedError


@stable_api
class HandlerContext(LoggerABC):
    """
    Context passed to handler methods for state related "things"
    """

    def __init__(
        self,
        resource: resources.Resource,
        dry_run: bool = False,
        action_id: Optional[uuid.UUID] = None,
        logger: Optional[logging.Logger] = None,
    ) -> None:
        self._resource = resource
        self._dry_run = dry_run
        self._cache: Dict[str, Any] = {}

        self._purged = False
        self._updated = False
        self._created = False
        self._change = const.Change.nochange

        self._changes: Dict[str, AttributeStateChange] = {}

        if action_id is None:
            action_id = uuid.uuid4()
        self._action_id = action_id
        self._status: Optional[ResourceState] = None
        self._logs: List[data.LogLine] = []
        self.logger: logging.Logger
        if logger is None:
            self.logger = LOGGER
        else:
            self.logger = logger

        self._facts: List[Dict[str, Any]] = []

    def set_fact(self, fact_id: str, value: str) -> None:
        """
        Send a fact to the Inmanta server.

        :param fact_id: The name of the fact.
        :param value: The actual value of the fact.
        """
        resource_id = self._resource.id.resource_str()
        fact = {
            "id": fact_id,
            "source": ParameterSource.fact.value,
            "value": value,
            "resource_id": resource_id,
        }
        self._facts.append(fact)

    @property
    def facts(self) -> List[Dict[str, Any]]:
        return self._facts

    @property
    def action_id(self) -> uuid.UUID:
        return self._action_id

    @property
    def status(self) -> Optional[const.ResourceState]:
        return self._status

    @property
    def logs(self) -> List[data.LogLine]:
        return self._logs

    def set_status(self, status: const.ResourceState) -> None:
        """
        Set the status of the handler operation.
        """
        self._status = status

    def is_dry_run(self) -> bool:
        """
        Is this a dryrun?
        """
        return self._dry_run

    def get(self, name: str) -> Any:
        return self._cache[name]

    def contains(self, key: str) -> bool:
        return key in self._cache

    def set(self, name: str, value: Any) -> None:
        self._cache[name] = value

    def set_created(self) -> None:
        self._created = True
        if self._change is not const.Change.nochange:
            raise InvalidOperation(f"Unable to set {const.Change.created} operation, {self._change} already set.")

        self._change = const.Change.created

    def set_purged(self) -> None:
        self._purged = True

        if self._change is not const.Change.nochange:
            raise InvalidOperation(f"Unable to set {const.Change.purged} operation, {self._change} already set.")

        self._change = const.Change.purged

    def set_updated(self) -> None:
        self._updated = True

        if self._change is not const.Change.nochange:
            raise InvalidOperation(f"Unable to set {const.Change.updated} operation, {self._change} already set.")

        self._change = const.Change.updated

    @property
    def changed(self) -> bool:
        return self._created or self._updated or self._purged

    @property
    def change(self) -> const.Change:
        return self._change

    def add_change(self, name: str, desired: object, current: object = None) -> None:
        """
        Report a change of a field. This field is added to the set of updated fields

        :param name: The name of the field that was updated
        :param desired: The desired value to which the field was updated (or should be updated)
        :param current: The value of the field before it was updated
        """
        self._changes[name] = AttributeStateChange(current=current, desired=desired)

    def add_changes(self, **kwargs: SimpleTypes) -> None:
        """
        Report a list of changes at once as kwargs

        :param key: The name of the field that was updated. This field is also added to the set of updated fields
        :param value: The desired value of the field.

        To report the previous value of the field, use the add_change method
        """
        for field, value in kwargs.items():
            self._changes[field] = AttributeStateChange(desired=value)

    def fields_updated(self, fields: str) -> None:
        """
        Report that fields have been updated
        """
        for field in fields:
            if field not in self._changes:
                self._changes[fields] = AttributeStateChange()

    @overload  # noqa: F811
    def update_changes(self, changes: Dict[str, AttributeStateChange]) -> None:
        pass

    @overload  # noqa: F811
    def update_changes(self, changes: Dict[str, Dict[str, Optional[SimpleTypes]]]) -> None:
        pass

    @overload  # noqa: F811
    def update_changes(self, changes: Dict[str, Tuple[SimpleTypes, SimpleTypes]]) -> None:
        pass

    def update_changes(  # noqa: F811
        self,
        changes: Union[
            Dict[str, AttributeStateChange],
            Dict[str, Dict[str, Optional[SimpleTypes]]],
            Dict[str, Tuple[SimpleTypes, SimpleTypes]],
        ],
    ) -> None:
        """
        Update the changes list with changes

        :param changes: This should be a dict with a value a dict containing "current" and "desired" keys
        """
        for attribute, change in changes.items():
            if isinstance(change, dict):
                self._changes[attribute] = AttributeStateChange(
                    current=change.get("current", None), desired=change.get("desired", None)
                )
            elif isinstance(change, tuple):
                if len(change) != 2:
                    raise InvalidOperation(
                        f"Reported changes for {attribute} not valid. Tuple changes should contain 2 element."
                    )
                self._changes[attribute] = AttributeStateChange(current=change[0], desired=change[1])
            elif isinstance(change, AttributeStateChange):
                self._changes[attribute] = change
            else:
                raise InvalidOperation(f"Reported changes for {attribute} not in a type that is recognized.")

    @property
    def changes(self) -> Dict[str, AttributeStateChange]:
        return self._changes

    def log_msg(self, level: int, msg: str, args: abc.Sequence[object], kwargs: abc.MutableMapping[str, object]) -> None:
        LOGGER.warning("Direct calls to the log_msg method are being deprecated, please use the LoggerABC interface instead.")
        self._log_msg(level, msg, *args, **kwargs)

    def _log_msg(self, level: int, msg: str, *args: object, exc_info: bool = False, **kwargs: object) -> None:
        if len(args) > 0:
            raise Exception("Args not supported")
        if exc_info:
            kwargs["traceback"] = traceback.format_exc()

        for k, v in kwargs.items():
            try:
                json_encode(v)
            except TypeError:
                if inmanta.RUNNING_TESTS:
                    # Fail the test when the value is not serializable
                    raise Exception(f"Failed to serialize argument for log message {k}={v}")
                else:
                    # In production, try to cast the non-serializable value to str to prevent the handler from failing.
                    kwargs[k] = str(v)

            except Exception as e:
                raise Exception("Exception during serializing log message arguments") from e
        log = data.LogLine.log(level, msg, **kwargs)
        self.logger.log(level, "resource %s: %s", self._resource.id.resource_version_str(), log._data["msg"], exc_info=exc_info)
        self._logs.append(log)


@stable_api
class ResourceHandler(object):
    """
    A baseclass for classes that handle resources. New handler are registered with the
    :func:`~inmanta.agent.handler.provider` decorator.

    The implementation of a handler should use the ``self._io`` instance to execute io operations. This io objects
    makes abstraction of local or remote operations. See :class:`~inmanta.agent.io.local.LocalIO` for the available
    operations.

    :param agent: The agent that is executing this handler.
    :param io: The io object to use.
    """

    def __init__(self, agent: "inmanta.agent.agent.AgentInstance", io: Optional["IOBase"] = None) -> None:
        self._agent = agent

        if io is None:
            raise Exception("Unsupported: no resource mgmt in RH")
        else:
            self._io = io

        self._client: Optional[protocol.SessionClient] = None
        # explicit ioloop reference, as we don't want the ioloop for the current thread, but the one for the agent
        self._ioloop = agent.process._io_loop

    def run_sync(self, func: typing.Callable[[], typing.Awaitable[T]]) -> T:
        """
        Run a the given async function on the ioloop of the agent. It will block the current thread until the future
        resolves.

        :param func: A function that returns a yieldable future.
        :return: The result of the async function.
        """
        f: Future[T] = Future()

        # This function is not typed because of generics, the used methods and currying
        def run() -> None:
            try:
                result = func()
                if result is not None:
                    from tornado.gen import convert_yielded

                    result = convert_yielded(result)
                    concurrent.chain_future(result, f)
            except Exception as e:
                f.set_exception(e)

        self._ioloop.call_soon_threadsafe(run)

        return f.result()

    def set_cache(self, cache: AgentCache) -> None:
        self.cache = cache

    def get_client(self) -> protocol.SessionClient:
        """
        Get the client instance that identifies itself with the agent session.

        :return: A client that is associated with the session of the agent that executes this handler.
        """
        if self._client is None:
            self._client = protocol.SessionClient("agent", self._agent.sessionid)
        return self._client

    def can_reload(self) -> bool:
        """
        Can this handler reload?

        :return: Return true if this handler needs to reload on requires changes.
        """
        return False

    def do_reload(self, ctx: HandlerContext, resource: resources.Resource) -> None:
        """
        Perform a reload of this resource.

        :param ctx: Context object to report changes and logs to the agent and server.
        :param resource: The resource to reload.
        """

    def pre(self, ctx: HandlerContext, resource: resources.Resource) -> None:
        """
        Method executed before a handler operation (Facts, dryrun, real deployment, ...) is executed. Override this method
        to run before an operation.

        :param ctx: Context object to report changes and logs to the agent and server.
        :param resource: The resource to query facts for.
        """

    def post(self, ctx: HandlerContext, resource: resources.Resource) -> None:
        """
        Method executed after an operation. Override this method to run after an operation.

        :param ctx: Context object to report changes and logs to the agent and server.
        :param resource: The resource to query facts for.
        """

    def close(self) -> None:
        pass

    def _diff(self, current: resources.Resource, desired: resources.Resource) -> typing.Dict[str, typing.Dict[str, typing.Any]]:
        """
        Calculate the diff between the current and desired resource state.

        :param current: The current state of the resource
        :param desired: The desired state of the resource
        :return: A dict with key the name of the field and value another dict with "current" and "desired" as keys for
                 fields that require changes.
        """
        changes = {}

        # check attributes
        for field in current.__class__.fields:
            current_value = getattr(current, field)
            desired_value = getattr(desired, field)

            if current_value != desired_value and desired_value is not None:
                changes[field] = {"current": current_value, "desired": desired_value}

        return changes

    def check_resource(self, ctx: HandlerContext, resource: resources.Resource) -> resources.Resource:
        """
        Check the current state of a resource

        :param ctx: Context object to report changes and logs to the agent and server.
        :param resource: The resource to check the current state of.
        :return: A resource to represents the current state. Use the :func:`~inmanta.resources.Resource.clone` to create
                 clone of the given resource that can be modified.
        """
        raise NotImplementedError()

    def list_changes(self, ctx: HandlerContext, resource: resources.Resource) -> typing.Dict[str, typing.Dict[str, typing.Any]]:
        """
        Returns the changes required to bring the resource on this system in the state described in the resource entry.
        This method calls :func:`~inmanta.agent.handler.ResourceHandler.check_resource`

        :param ctx: Context object to report changes and logs to the agent and server.
        :param resource: The resource to check the current state of.
        :return: A dict with key the name of the field and value another dict with "current" and "desired" as keys for
                 fields that require changes.
        """
        current = self.check_resource(ctx, resource)
        return self._diff(current, resource)

    def do_changes(self, ctx: HandlerContext, resource: resources.Resource, changes: Dict[str, Dict[str, object]]) -> None:
        """
        Do the changes required to bring the resource on this system in the state of the given resource.

        :param ctx: Context object to report changes and logs to the agent and server.
        :param resource: The resource to check the current state of.
        :param changes: The changes that need to occur as reported by
                        :func:`~inmanta.agent.handler.ResourceHandler.list_changes`
        """
        raise NotImplementedError()

    def deploy(
        self,
        ctx: HandlerContext,
        resource: resources.Resource,
        requires: Dict[ResourceIdStr, ResourceState],
    ) -> None:
        """
        This method is always be called by the agent, even when one of the requires of the given resource
        failed to deploy. The default implementation of this method will deploy the given resource when all its
        requires were deployed successfully. Override this method if a different condition determines whether the
        resource should deploy.

        :param ctx: Context object to report changes and logs to the agent and server.
        :param resource: The resource to deploy
        :param requires: A dictionary mapping the resource id of each dependency of the given resource to its resource state.
        """

        def _call_resource_did_dependency_change() -> typing.Awaitable[Result]:
            return self.get_client().resource_did_dependency_change(
                tid=self._agent.environment, rvid=resource.id.resource_version_str()
            )

        def _should_reload() -> bool:
            if not self.can_reload():
                return False
            result = self.run_sync(_call_resource_did_dependency_change)
            if not result.result:
                raise Exception("Failed to determine whether resource should reload")

            if result.code != 200:
                error_msg_from_server = f": {result.result['message']}" if "message" in result.result else ""
                raise Exception(f"Failed to determine whether resource should reload{error_msg_from_server}")
            return result.result["data"]

        def filter_resources_in_unexpected_state(
            reqs: Dict[ResourceIdStr, ResourceState]
        ) -> Dict[ResourceIdStr, ResourceState]:
            """
            Return a sub-dictionary of reqs with only those resources that are in an unexpected state.
            """
            unexpected_states = {
                const.ResourceState.available,
                const.ResourceState.dry,
                const.ResourceState.undefined,
                const.ResourceState.skipped_for_undefined,
                const.ResourceState.deploying,
            }
            return {rid: state for rid, state in reqs.items() if state in unexpected_states}

        resources_in_unexpected_state = filter_resources_in_unexpected_state(requires)
        if resources_in_unexpected_state:
            ctx.set_status(const.ResourceState.skipped)
            ctx.warning(
                "Resource %(resource)s skipped because a dependency is in an unexpected state: %(unexpected_states)s",
                resource=resource.id.resource_version_str(),
                unexpected_states=str({rid: state.value for rid, state in resources_in_unexpected_state.items()}),
            )
            return

        failed_dependencies = [req for req, status in requires.items() if status != ResourceState.deployed]
        if not any(failed_dependencies):
            self.execute(ctx, resource)
            if _should_reload():
                self.do_reload(ctx, resource)
        else:
            ctx.set_status(const.ResourceState.skipped)
            ctx.info(
                "Resource %(resource)s skipped due to failed dependencies: %(failed)s",
                resource=resource.id.resource_version_str(),
                failed=str(failed_dependencies),
            )

    def execute(self, ctx: HandlerContext, resource: resources.Resource, dry_run: bool = False) -> None:
        """
        Update the given resource. This method is called by the agent. Most handlers will not override this method
        and will only override :func:`~inmanta.agent.handler.ResourceHandler.check_resource`, optionally
        :func:`~inmanta.agent.handler.ResourceHandler.list_changes` and
        :func:`~inmanta.agent.handler.ResourceHandler.do_changes`

        :param ctx: Context object to report changes and logs to the agent and server.
        :param resource: The resource to check the current state of.
        :param dry_run: True will only determine the required changes but will not execute them.
        """
        try:
            self.pre(ctx, resource)

            changes = self.list_changes(ctx, resource)
            ctx.update_changes(changes)

            if not dry_run:
                self.do_changes(ctx, resource, changes)
                ctx.set_status(const.ResourceState.deployed)
            else:
                ctx.set_status(const.ResourceState.dry)
        except SkipResource as e:
            ctx.set_status(const.ResourceState.skipped)
            ctx.warning(msg="Resource %(resource_id)s was skipped: %(reason)s", resource_id=resource.id, reason=e.args)

        except Exception as e:
            ctx.set_status(const.ResourceState.failed)
            ctx.exception(
                "An error occurred during deployment of %(resource_id)s (exception: %(exception)s",
                resource_id=resource.id,
                exception=f"{e.__class__.__name__}('{e}')",
            )
        finally:
            try:
                self.post(ctx, resource)
            except Exception as e:
                ctx.exception(
                    "An error occurred after deployment of %(resource_id)s (exception: %(exception)s",
                    resource_id=resource.id,
                    exception=f"{e.__class__.__name__}('{e}')",
                )

    def facts(self, ctx: HandlerContext, resource: resources.Resource) -> Dict[str, object]:
        """
        Override this method to implement fact querying. A queried fact can be reported back in two different ways:
        either via the return value of this method or by adding the fact to the HandlerContext via the
        :func:`~inmanta.agent.handler.HandlerContext.set_fact` method. :func:`~inmanta.agent.handler.ResourceHandler.pre`
        and :func:`~inmanta.agent.handler.ResourceHandler.post` are called before and after this method.

        :param ctx: Context object to report changes, logs and facts to the agent and server.
        :param resource: The resource to query facts for.
        :return: A dict with fact names as keys and facts values.
        """
        return {}

    def check_facts(self, ctx: HandlerContext, resource: resources.Resource) -> Dict[str, object]:
        """
        This method is called by the agent to query for facts. It runs :func:`~inmanta.agent.handler.ResourceHandler.pre`
        and :func:`~inmanta.agent.handler.ResourceHandler.post`. This method calls
        :func:`~inmanta.agent.handler.ResourceHandler.facts` to do the actually querying.

        :param ctx: Context object to report changes and logs to the agent and server.
        :param resource: The resource to query facts for.
        :return: A dict with fact names as keys and facts values.
        """
        try:
            self.pre(ctx, resource)
            facts = self.facts(ctx, resource)
        finally:
            try:
                self.post(ctx, resource)
            except Exception as e:
                ctx.exception(
                    "An error occurred after getting facts about %(resource_id)s (exception: %(exception)s",
                    resource_id=resource.id,
                    exception=f"{e.__class__.__name__}('{e}')",
                )

        return facts

    def available(self, resource: resources.Resource) -> bool:
        """
        Returns true if this handler is available for the given resource

        :param resource: Is this handler available for the given resource?
        :return: Available or not?
        """
        return True

    def get_file(self, hash_id: str) -> Optional[bytes]:
        """
        Retrieve a file from the fileserver identified with the given id. The convention is to use the sha1sum of the
        content to identify it.

        :param hash_id: The id of the content/file to retrieve from the server.
        :return: The content in the form of a bytestring or none is the content does not exist.
        """

        def call() -> typing.Awaitable[Result]:
            return self.get_client().get_file(hash_id)

        result = self.run_sync(call)
        if result.code == 404:
            return None
        elif result.result and result.code == 200:
            file_contents = base64.b64decode(result.result["content"])
            actual_hash_of_file = hash_file(file_contents)
            if hash_id != actual_hash_of_file:
                raise Exception(f"File hash verification failed, expected: {hash_id} but got {actual_hash_of_file}")
            return file_contents
        else:
            raise Exception("An error occurred while retrieving file %s" % hash_id)

    def stat_file(self, hash_id: str) -> bool:
        """
        Check if a file exists on the server. This method does and async call to the server and blocks on the result.

        :param hash_id: The id of the file on the server. The convention is the use the sha1sum of the content as id.
        :return: True if the file is available on the server.
        """

        def call() -> typing.Awaitable[Result]:
            return self.get_client().stat_file(hash_id)

        result = self.run_sync(call)
        return result.code == 200

    def upload_file(self, hash_id: str, content: bytes) -> None:
        """
        Upload a file to the server

        :param hash_id: The id to identify the content. The convention is to use the sha1sum of the content to identify it.
        :param content: A byte string with the content
        """

        def call() -> typing.Awaitable[Result]:
            return self.get_client().upload_file(id=hash_id, content=base64.b64encode(content).decode("ascii"))

        try:
            self.run_sync(call)
        except Exception:
            raise Exception("Unable to upload file to the server.")


@stable_api
class CRUDHandler(ResourceHandler):
    """
    This handler base class requires CRUD methods to be implemented: create, read, update and delete. Such a handler
    only works on purgeable resources.
    """

    def read_resource(self, ctx: HandlerContext, resource: resources.PurgeableResource) -> None:
        """
        This method reads the current state of the resource. It provides a copy of the resource that should be deployed,
        the method implementation should modify the attributes of this resource to the current state.

        :param ctx: Context can be used to pass value discovered in the read method to the CUD methods. For example, the
                   id used in API calls
        :param resource: A clone of the desired resource state. The read method need to set values on this object.
        :raise SkipResource: Raise this exception when the handler should skip this resource
        :raise ResourcePurged: Raise this exception when the resource does not exist yet.
        """

    def create_resource(self, ctx: HandlerContext, resource: resources.PurgeableResource) -> None:
        """
        This method is called by the handler when the resource should be created.

        :param context: Context can be used to get values discovered in the read method. For example, the id used in API
                        calls. This context should also be used to let the handler know what changes were made to the
                        resource.
        :param resource: The desired resource state.
        """

    def delete_resource(self, ctx: HandlerContext, resource: resources.PurgeableResource) -> None:
        """
        This method is called by the handler when the resource should be deleted.

        :param ctx: Context can be used to get values discovered in the read method. For example, the id used in API
                    calls. This context should also be used to let the handler know what changes were made to the
                    resource.
        :param resource: The desired resource state.
        """

    def update_resource(
        self, ctx: HandlerContext, changes: Dict[str, Dict[str, Any]], resource: resources.PurgeableResource
    ) -> None:
        """
        This method is called by the handler when the resource should be updated.

        :param ctx: Context can be used to get values discovered in the read method. For example, the id used in API
                    calls. This context should also be used to let the handler know what changes were made to the
                    resource.
        :param changes: A map of resource attributes that should be changed. Each value is a tuple with the current and the
                        desired value.
        :param resource: The desired resource state.
        """

    def calculate_diff(
        self, ctx: HandlerContext, current: resources.Resource, desired: resources.Resource
    ) -> typing.Dict[str, typing.Dict[str, typing.Any]]:
        """
        Calculate the diff between the current and desired resource state.

        :param ctx: Context can be used to get values discovered in the read method. For example, the id used in API
                    calls. This context should also be used to let the handler know what changes were made to the
                    resource.
        :param current: The current state of the resource
        :param desired: The desired state of the resource
        :return: A dict with key the name of the field and value another dict with "current" and "desired" as keys for
                 fields that require changes.
        """
        return self._diff(current, desired)

    def execute(self, ctx: HandlerContext, resource: resources.Resource, dry_run: Optional[bool] = None) -> None:
        """
        Update the given resource. This method is called by the agent. Override the CRUD methods of this class.

        :param ctx: Context object to report changes and logs to the agent and server.
        :param resource: The resource to check the current state of.
        :param dry_run: True will only determine the required changes but will not execute them.
        """
        try:
            self.pre(ctx, resource)

            # current is clone, except for purged is set to false to prevent a bug that occurs often where the desired
            # state defines purged=true but the read_resource fails to set it to false if the resource does exist
            desired = resource
            assert isinstance(desired, resources.PurgeableResource)
            current = desired.clone(purged=False)
            assert isinstance(current, resources.PurgeableResource)
            changes: typing.Dict[str, typing.Dict[str, typing.Any]] = {}
            try:
                ctx.debug("Calling read_resource")
                self.read_resource(ctx, current)
                changes = self.calculate_diff(ctx, current, desired)

            except ResourcePurged:
                if not desired.purged:
                    changes["purged"] = dict(desired=desired.purged, current=True)

            for field, values in changes.items():
                ctx.add_change(field, desired=values["desired"], current=values["current"])

            if not dry_run:
                if "purged" in changes:
                    if not changes["purged"]["desired"]:
                        ctx.debug("Calling create_resource")
                        self.create_resource(ctx, desired)
                    else:
                        ctx.debug("Calling delete_resource")
                        self.delete_resource(ctx, desired)

                elif not desired.purged and len(changes) > 0:
                    ctx.debug("Calling update_resource", changes=changes)
                    self.update_resource(ctx, dict(changes), desired)

                ctx.set_status(const.ResourceState.deployed)
            else:
                ctx.set_status(const.ResourceState.dry)

        except SkipResource as e:
            ctx.set_status(const.ResourceState.skipped)
            ctx.warning(msg="Resource %(resource_id)s was skipped: %(reason)s", resource_id=resource.id, reason=e.args)

        except Exception as e:
            ctx.set_status(const.ResourceState.failed)
            ctx.exception(
                "An error occurred during deployment of %(resource_id)s (exception: %(exception)s)",
                resource_id=resource.id,
                exception=f"{e.__class__.__name__}('{e}')",
                traceback=traceback.format_exc(),
            )
        finally:
            try:
                self.post(ctx, resource)
            except Exception as e:
                ctx.exception(
                    "An error occurred after deployment of %(resource_id)s (exception: %(exception)s",
                    resource_id=resource.id,
                    exception=f"{e.__class__.__name__}('{e}')",
                )


TPurgeableResource = TypeVar("TPurgeableResource", bound=resources.PurgeableResource)


@stable_api
class CRUDHandlerGeneric(CRUDHandler, Generic[TPurgeableResource]):
    """
    This class offers the same functionality as the CRUDHandler class, but was made generic on the type of PurgeableResource.
    """

    def read_resource(self, ctx: HandlerContext, resource: TPurgeableResource) -> None:
        pass

    def create_resource(self, ctx: HandlerContext, resource: TPurgeableResource) -> None:
        pass

    def delete_resource(self, ctx: HandlerContext, resource: TPurgeableResource) -> None:
        pass

    def update_resource(self, ctx: HandlerContext, changes: Dict[str, Dict[str, Any]], resource: TPurgeableResource) -> None:
        pass

    def calculate_diff(
        self, ctx: HandlerContext, current: TPurgeableResource, desired: TPurgeableResource
    ) -> typing.Dict[str, typing.Dict[str, typing.Any]]:
        return super().calculate_diff(ctx, current, desired)

    def execute(self, ctx: HandlerContext, resource: TPurgeableResource, dry_run: bool = False) -> None:
        super().execute(ctx, resource, dry_run)


class Commander(object):
    """
    This class handles commands
    """

    __command_functions: Dict[str, Dict[str, Type[ResourceHandler]]] = defaultdict(dict)

    @classmethod
    def get_handlers(cls) -> Dict[str, Dict[str, Type[ResourceHandler]]]:
        return cls.__command_functions

    @classmethod
    def reset(cls) -> None:
        cls.__command_functions = defaultdict(dict)

    @classmethod
    def close(cls) -> None:
        pass

    @classmethod
    def _get_instance(
        cls, handler_class: Type[ResourceHandler], agent: "inmanta.agent.agent.AgentInstance", io: "IOBase"
    ) -> ResourceHandler:
        new_instance = handler_class(agent, io)
        return new_instance

    @classmethod
    def get_provider(
        cls, cache: AgentCache, agent: "inmanta.agent.agent.AgentInstance", resource: resources.Resource
    ) -> ResourceHandler:
        """
        Return a provider to handle the given resource
        """
        resource_id = resource.id
        resource_type = resource_id.get_entity_type()
        try:
            agent_io = io.get_io(cache, agent.uri, resource_id.get_version())
        except Exception:
            LOGGER.exception("Exception raised during creation of IO for uri %s", agent.uri)
            raise Exception("No handler available for %s (no io available)" % resource_id)

        if agent_io is None:
            # Skip this resource
            raise Exception("No handler available for %s (no io available)" % resource_id)

        available = []
        if resource_type in cls.__command_functions:
            for handlr in cls.__command_functions[resource_type].values():
                h = cls._get_instance(handlr, agent, agent_io)
                if h.available(resource):
                    available.append(h)
                else:
                    h.close()

        if len(available) > 1:
            for h in available:
                h.close()

            agent_io.close()
            raise Exception("More than one handler selected for resource %s" % resource.id)

        elif len(available) == 1:
            return available[0]

        raise Exception("No resource handler registered for resource of type %s" % resource_type)

    @classmethod
    def add_provider(cls, resource: str, name: str, provider: Type["ResourceHandler"]) -> None:
        """
        Register a new provider

        :param resource: the name of the resource this handler applies to
        :param name: the name of the handler itself
        :param provider: the handler function
        """
        if resource in cls.__command_functions and name in cls.__command_functions[resource]:
            del cls.__command_functions[resource][name]

        cls.__command_functions[resource][name] = provider

    @classmethod
    def get_providers(cls) -> typing.Iterator[Tuple[str, typing.Type["ResourceHandler"]]]:
        """Return an iterator over resource type, handler definition"""
        for resource_type, handler_map in cls.__command_functions.items():
            for handle_name, handler_class in handler_map.items():
                yield (resource_type, handler_class)

    @classmethod
    def get_provider_class(cls, resource_type: str, name: str) -> Optional[typing.Type["ResourceHandler"]]:
        """
        Return the class of the handler for the given type and with the given name
        """
        if resource_type not in cls.__command_functions:
            return None

        if name not in cls.__command_functions[resource_type]:
            return None

        return cls.__command_functions[resource_type][name]


class HandlerNotAvailableException(Exception):
    """
    This exception is thrown when a resource handler cannot perform its job. For example, the admin interface
    it connects to is not available.
    """


@stable_api
class PythonLogger(LoggerABC):
    """
    This class implements the LoggerABC interface and is a standalone wrapper around a logging.Logger.
    """

    def __init__(self, logger: logging.Logger) -> None:
        self.logger = logger

    def _log_msg(
        self,
        level: int,
        msg: str,
        *args: object,
        exc_info: bool = False,
        **kwargs: object,
    ) -> None:
        if len(args) > 0:
            raise Exception("Args not supported")

        if kwargs:
            self.logger.log(level, msg, kwargs, exc_info=exc_info)
        else:
            self.logger.log(level, msg, exc_info=exc_info)



# Copyright 2019 Inmanta
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Contact: code@inmanta.com

"""
The protocol module contains all the code related handle communication between Inmanta components. This documentation
describes its structure. Communication is performed with a central server component. Clients connect to this central
server either with a normal call or with a calls over a session. Session based clients can communicate also receive
calls from the server.

.. uml:: structure.puml
    :align: center


AgentEndpoint and Server, Session and SessionManager implement a reverse channel using a heartbeat mechanism and a
queue of futures for the Server to call methods on the Agent. The Server queues up requests and returns them in a
agent heartbeat. The AgentEndpoint process the returned calls and executes them with RESTBase._execute_call using
the same semantics as the Server for a normal http call.

Generic protocol code structure
-------------------------------

    * common: Classes and helper methods used by various protocol modules. The logic to transform a function to a REST
              call is located in this module. Mainly in :class:`~inmanta.protocol.common.MethodProperties`
    * decorators: Contains the decorator to add API method calls to endpoints and the decorator to register handlers
                  on an endpoint.
    * endpoints: Endpoint base class and multiple types of client endpoints. The server endpoint is defined in
                 inmanta.server.protocol
    * exceptions: Excpetions to signal errors over REST.
    * methods: This module defines the signature of each available API call handled with protocol

    * inmanta.server.protocol contains the session management and the logic required to communicate in the reverse direction
      over a session.

REST specific
------------------
All communication is over REST (including two-way communication). The rest package contains all the code related
to REST transport with Tornado, together with the code in :module:`~inmanta.server.protocol`

"""

# flake8: noqa: F401, F403

from . import methods, methods_v2
from .common import Response, Result, decode_token, encode_token, gzipped_json, json_encode
from .decorators import handle, method, typedmethod
from .endpoints import Client, SessionClient, SessionEndpoint, SyncClient, VersionMatch

__all__ = [
    "Response",
    "Result",
    "decode_token",
    "encode_token",
    "gzipped_json",
    "json_encode",
    "Client",
    "SessionClient",
    "SessionEndpoint",
    "SyncClient",
    "VersionMatch",
    "handle",
    "method",
    "typedmethod",
]



"""
    Copyright 2022 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import abc
import json
from abc import ABC
from datetime import datetime
from typing import Dict, Generic, List, Optional, Sequence, Tuple, Type, TypeVar, Union, cast
from urllib import parse
from urllib.parse import quote
from uuid import UUID

from asyncpg import Record

from inmanta import data
from inmanta.data import (
    APILIMIT,
    PRIMITIVE_SQL_TYPES,
    Agent,
    AgentOrder,
    CompileReportOrder,
    ConfigurationModel,
    DatabaseOrderV2,
    DesiredStateVersionOrder,
    DiscoveredResourceOrder,
    FactOrder,
    InvalidQueryParameter,
    InvalidSort,
    Notification,
    NotificationOrder,
    PagingOrder,
    Parameter,
    ParameterOrder,
    QueryFilter,
    Resource,
    ResourceAction,
    ResourceHistoryOrder,
    ResourceLogOrder,
    ResourceOrder,
    SimpleQueryBuilder,
    VersionedResourceOrder,
    model,
)
from inmanta.data.model import (
    BaseModel,
    CompileReport,
    DesiredStateLabel,
    DesiredStateVersion,
    Fact,
    LatestReleasedResource,
    PagingBoundaries,
    ResourceHistory,
    ResourceIdStr,
    ResourceLog,
    ResourceVersionIdStr,
)
from inmanta.protocol.exceptions import BadRequest
from inmanta.protocol.return_value_meta import ReturnValueWithMeta
from inmanta.resources import Id
from inmanta.server.validate_filter import (
    BooleanEqualityFilter,
    BooleanIsNotNullFilter,
    CombinedContainsFilterResourceState,
    ContainsFilter,
    ContainsFilterResourceAction,
    ContainsPartialFilter,
    DateRangeFilter,
    Filter,
    FilterValidator,
    IntRangeFilter,
    InvalidFilter,
    LogLevelFilter,
)
from inmanta.types import JsonType, SimpleTypes
from inmanta.util import datetime_utc_isoformat

T_ORDER = TypeVar("T_ORDER", bound=DatabaseOrderV2)
T_DTO = TypeVar("T_DTO", bound=BaseModel)


class RequestedPagingBoundaries:
    """Represents the lower and upper bounds that the user requested for the paging boundaries, if any."""

    def __init__(
        self,
        start: Optional[PRIMITIVE_SQL_TYPES],
        end: Optional[PRIMITIVE_SQL_TYPES],
        first_id: Optional[PRIMITIVE_SQL_TYPES],
        last_id: Optional[PRIMITIVE_SQL_TYPES],
    ) -> None:
        self.start = start
        self.end = end
        self.first_id = first_id
        self.last_id = last_id

    def validate(
        self,
    ) -> None:
        start = self.start
        end = self.end
        first_id = self.first_id
        last_id = self.last_id
        if (start is not None) and (end is not None):
            raise InvalidQueryParameter(
                f"Only one of start and end parameters is allowed at the same time. Received start: {start}, end: {end}"
            )
        if (first_id is not None) and (last_id is not None):
            raise InvalidQueryParameter(
                f"Only one of first_id and last_id parameters is allowed at the same time. "
                f"Received first_id: {first_id}, last_id: {last_id}"
            )

        if self.has_start() and self.has_end():
            raise InvalidQueryParameter(
                f"Start and end parameters can not be set at the same time: "
                f"Received first_id: {first_id}, last_id: {last_id}, start: {start}, end: {end}"
            )

    def has_start(self) -> bool:
        return (self.start is not None) or (self.first_id is not None)

    def has_end(self) -> bool:
        return (self.end is not None) or (self.last_id is not None)


class PagingMetadata:
    def __init__(self, total: int, before: int, after: int, page_size: int) -> None:
        self.total = total
        self.before = before
        self.after = after
        self.page_size = page_size

    def to_dict(self) -> Dict[str, int]:
        return {
            "total": self.total,
            "before": self.before,
            "after": self.after,
            "page_size": self.page_size,
        }


class DataView(FilterValidator, Generic[T_ORDER, T_DTO], ABC):
    def __init__(
        self,
        order: T_ORDER,
        limit: Optional[int] = None,
        first_id: Optional[PRIMITIVE_SQL_TYPES] = None,
        last_id: Optional[PRIMITIVE_SQL_TYPES] = None,
        start: Optional[PRIMITIVE_SQL_TYPES] = None,
        end: Optional[PRIMITIVE_SQL_TYPES] = None,
        filter: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        self.limit = self.validate_limit(limit)
        self.raw_filter = filter or {}
        self.filter: Dict[str, QueryFilter] = self.process_filters(filter)
        self.order = order
        self.requested_page_boundaries = RequestedPagingBoundaries(start, end, first_id, last_id)
        self.requested_page_boundaries.validate()

    @abc.abstractmethod
    def get_base_url(self) -> str:
        """
        Return the base URL used to construct the paging links

        e.g. "/api/v2/resource"
        """
        pass

    def get_extra_url_parameters(self) -> Dict[str, str]:
        """
        Return additional URL query parameters required to construct the paging links
        """
        return {}

    async def get_data(self) -> Tuple[Sequence[T_DTO], Optional[PagingBoundaries]]:
        query_builder = self.get_base_query()

        # In this method, we use `data.Resource`
        # we need the generic functionality of `data.BaseDocument`
        # But that one doesn't actually hold a connection, as it is abstract

        # Project
        query_builder = query_builder.filter(
            *data.Resource.get_composed_filter_with_query_types(
                offset=query_builder.offset, col_name_prefix=None, **self.filter
            )
        )
        query_builder = self.clip_to_page(query_builder)
        sql_query, values = query_builder.build()

        records = await data.Resource.select_query(sql_query, values, no_obj=True)
        dtos = self.construct_dtos(records)

        paging_boundaries = None
        if dtos:
            paging_boundaries = self.order.get_paging_boundaries(dict(records[0]), dict(records[-1]))
        return dtos, paging_boundaries

    @abc.abstractmethod
    def construct_dtos(self, records: Sequence[Record]) -> Sequence[T_DTO]:
        """
        Convert the sequence of records into a sequence of DTO's
        """
        pass

    @abc.abstractmethod
    def get_base_query(self) -> SimpleQueryBuilder:
        """
        Return the base query to get the data.

        Must contain select, from and where clause if specific filtering is required
        """
        pass

    @property
    @abc.abstractmethod
    def allowed_filters(self) -> Dict[str, Type[Filter]]:
        """
        Return the specification of the allowed filters, see FilterValidator
        """
        pass

    def clip_to_page(self, query_builder: SimpleQueryBuilder) -> SimpleQueryBuilder:
        """
        Update the query builder to constrain it to the page boundaries, order and size
        """
        order = self.order.get_order()
        backward_paging: bool = (order == PagingOrder.ASC and self.requested_page_boundaries.has_end()) or (
            order == PagingOrder.DESC and self.requested_page_boundaries.has_start()
        )

        return (
            query_builder.filter(
                *self.order.as_filter(
                    query_builder.offset,
                    self.requested_page_boundaries.start,
                    self.requested_page_boundaries.first_id,
                    start=True,
                )
            )
            .filter(
                *self.order.as_filter(
                    query_builder.offset,
                    self.requested_page_boundaries.end,
                    self.requested_page_boundaries.last_id,
                    start=False,
                )
            )
            .order_and_limit(self.order, self.limit, backward_paging)
        )

    async def execute(self) -> ReturnValueWithMeta[Sequence[T_DTO]]:
        """
        Main entry point:
        - get the data from the DB
        - get page counts
        - format paging links

        :return: Complete API ReturnValueWithMeta ready to go out
        """
        try:
            dtos, paging_boundaries_in = await self.get_data()

            paging_boundaries: Union[PagingBoundaries, RequestedPagingBoundaries]
            if paging_boundaries_in:
                paging_boundaries = paging_boundaries_in
            else:
                # nothing found now, use the current page boundaries to determine if something exists before us
                paging_boundaries = self.requested_page_boundaries

            metadata = await self._get_page_count(paging_boundaries)
            links = await self.prepare_paging_links(dtos, paging_boundaries, metadata)
            return ReturnValueWithMeta(response=dtos, links=links if links else {}, metadata=metadata.to_dict())
        except (InvalidFilter, InvalidSort, data.InvalidQueryParameter, data.InvalidFieldNameException) as e:
            raise BadRequest(e.message) from e

    # Paging helpers

    async def _get_page_count(self, bounds: Union[PagingBoundaries, RequestedPagingBoundaries]) -> PagingMetadata:
        """
        Construct the page counts,

        either from the PagingBoundaries if we have a valid page,
        or from the RequestedPagingBoundaries if we got an empty page
        """
        query_builder = self.get_base_query()

        query_builder = query_builder.filter(
            *data.BaseDocument.get_composed_filter_with_query_types(
                offset=query_builder.offset, col_name_prefix=None, **self.filter
            )
        )

        order = self.order.get_order()
        reversed = order == PagingOrder.DESC

        end, start, last_id, first_id = bounds.end, bounds.start, bounds.last_id, bounds.first_id

        if reversed:
            end, start = start, end
            last_id, first_id = first_id, last_id

        values = query_builder.values
        after_filter_statements, after_values = self.order.as_filter(len(values) + 1, start, first_id, start=not reversed)
        values.extend(after_values)
        after_filter = data.BaseDocument._join_filter_statements(after_filter_statements)

        before_filter_statements, before_values = self.order.as_filter(len(values) + 1, end, last_id, start=reversed)
        values.extend(before_values)
        before_filter = data.BaseDocument._join_filter_statements(before_filter_statements)

        # If the currently requested page was empty,
        # we are working from RequestedPagingBoundaries instead of PagingBoundaries
        # If the RequestedPagingBoundaries has only nulls, we are not paging and the total count is 0
        # If the RequestedPagingBoundaries has one pair of nulls,
        #   we are paging but the current page is empty and so are the next ones
        # the `as_filter` method will return empty if the input is null
        # so we can conclude that the value for an empty filter is always 0

        if not before_filter and not after_filter:
            # Table is empty
            # Don't even make the query
            return PagingMetadata(
                total=0,
                before=0,
                after=0,
                page_size=self.limit,
            )

        select_clause = (
            "SELECT COUNT(*) as count_total"
            + (f", COUNT(*) filter ({before_filter}) as count_before" if before_filter else "")
            + (f", COUNT(*) filter ({after_filter}) as count_after " if after_filter else "")
        )

        query_builder = query_builder.select(select_clause)

        sql_query, values = query_builder.build()
        result = await data.Resource.select_query(sql_query, values, no_obj=True)
        result = cast(List[Record], result)
        if not result:
            raise InvalidQueryParameter("Could not determine page bounds")
        return PagingMetadata(
            total=cast(int, result[0]["count_total"]),
            before=cast(int, result[0].get("count_before", 0)),
            after=cast(int, result[0].get("count_after", 0)),
            page_size=self.limit,
        )

    async def prepare_paging_links(
        self,
        dtos: Sequence[T_DTO],
        paging_boundaries: Union[PagingBoundaries, RequestedPagingBoundaries],
        meta: PagingMetadata,
    ) -> Dict[str, str]:
        """
        Construct the paging links
        """
        links = {}

        url_query_params: Dict[str, Optional[Union[SimpleTypes, List[str]]]] = {
            "limit": self.limit,
            "sort": str(self.order),
        }

        for key, value in self.raw_filter.items():
            if value is not None:
                url_query_params[f"filter.{key}"] = value

        url_query_params.update(self.get_extra_url_parameters())

        if dtos:
            base_url = self.get_base_url()

            def value_to_string(value: Union[str, int, UUID, datetime]) -> str:
                if isinstance(value, datetime):
                    # Accross API boundaries, all naive datetime instances are assumed UTC.
                    # Returns ISO timestamp implicitly in UTC.
                    return datetime_utc_isoformat(value, naive_utc=True)
                return str(value)

            def make_link(**args: Optional[Union[str, int, UUID, datetime]]) -> str:
                params = url_query_params.copy()
                params.update({k: value_to_string(v) for k, v in args.items() if v is not None})
                return f"{base_url}?{parse.urlencode(params, doseq=True)}"

            link_with_end = make_link(
                end=paging_boundaries.end,
                last_id=paging_boundaries.last_id,
            )
            link_with_start = make_link(
                start=paging_boundaries.start,
                first_id=paging_boundaries.first_id,
            )

            has_next = meta.after > 0
            if has_next:
                if self.order.get_order() == "DESC":
                    links["next"] = link_with_end
                else:
                    links["next"] = link_with_start

            has_prev = meta.before > 0
            if has_prev:
                if self.order.get_order() == "DESC":
                    links["prev"] = link_with_start
                else:
                    links["prev"] = link_with_end
                # First page
                links["first"] = make_link()

            # Same page
            links["self"] = make_link(
                first_id=self.requested_page_boundaries.first_id,
                start=self.requested_page_boundaries.start,
            )
            # TODO: last links
        return links

    def validate_limit(self, limit: Optional[int]) -> int:
        if limit is None:
            return APILIMIT
        if limit > APILIMIT:
            raise BadRequest(f"limit parameter can not exceed {APILIMIT}, got {limit}.")
        return limit


class ResourceView(DataView[ResourceOrder, model.LatestReleasedResource]):
    def __init__(
        self,
        env: data.Environment,
        limit: Optional[int] = None,
        first_id: Optional[ResourceVersionIdStr] = None,
        last_id: Optional[ResourceVersionIdStr] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        filter: Optional[Dict[str, List[str]]] = None,
        sort: str = "resource_type.desc",
        deploy_summary: bool = False,
    ) -> None:
        super().__init__(
            order=ResourceOrder.parse_from_string(sort),
            limit=limit,
            first_id=first_id,
            last_id=last_id,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = env
        self.deploy_summary = deploy_summary

    @property
    def allowed_filters(self) -> Dict[str, Type[Filter]]:
        return {
            "resource_type": ContainsPartialFilter,
            "agent": ContainsPartialFilter,
            "resource_id_value": ContainsPartialFilter,
            "status": CombinedContainsFilterResourceState,
        }

    def get_base_url(self) -> str:
        return "/api/v2/resource"

    def get_extra_url_parameters(self) -> Dict[str, str]:
        return {"deploy_summary": str(self.deploy_summary)}

    def get_base_query(self) -> SimpleQueryBuilder:
        def subquery_latest_version_for_single_resource(higher_than: Optional[str]) -> str:
            """
            Returns a subquery to select a single row from a resource table:
                - for the first resource id higher than the given boundary
                - the highest (released) version for which a resource with this id exists

            :param higher_than: If given, the subquery selects the first resource id higher than this value (may be a column).
                If not given, the subquery selects the first resource id, period.
            """
            higher_than_condition: str = f"AND r.resource_id > {higher_than}" if higher_than is not None else ""
            return f"""
                SELECT
                    r.resource_id,
                    r.attributes,
                    r.resource_type,
                    r.agent,
                    r.resource_id_value,
                    r.model,
                    r.environment,
                    (
                        CASE WHEN (SELECT r.model < latest_version.version FROM latest_version)
                            THEN 'orphaned' -- use the CTE to check the status
                            ELSE r.status::text
                        END
                    ) as status
                FROM resource r
                JOIN configurationmodel cm ON r.model = cm.version AND r.environment = cm.environment
                WHERE r.environment = $1 AND cm.released = TRUE {higher_than_condition}
                ORDER BY resource_id, model DESC
                LIMIT 1
            """

        query_builder = SimpleQueryBuilder(
            select_clause="SELECT *",
            prelude=f"""
                /* the recursive CTE is the second one, but it has to be specified after 'WITH' if any of them are recursive */
                /* The latest_version CTE finds the maximum released version number in the environment */
                WITH RECURSIVE latest_version AS (
                    SELECT MAX(public.configurationmodel.version) as version
                    FROM public.configurationmodel
                    WHERE public.configurationmodel.released=TRUE AND environment=$1
                ),
                /*
                emulate a loose (or skip) index scan (https://wiki.postgresql.org/wiki/Loose_indexscan):
                1 resource_id at a time, select the latest (released) version it exists in.
                */
                cte AS (
                    /* Initial row for recursion: select relevant version for first resource */
                    ( {subquery_latest_version_for_single_resource(higher_than=None)} )
                    UNION ALL
                    SELECT next_r.*
                    FROM cte curr_r
                    CROSS JOIN LATERAL (
                        /* Recurse: select relevant version for next resource (one higher in the sort order than current) */
                        {subquery_latest_version_for_single_resource(higher_than="curr_r.resource_id")}
                    ) next_r
                )
            """,
            from_clause="FROM cte r",
            values=[self.environment.id],
        )
        return query_builder

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[model.LatestReleasedResource]:
        dtos: Sequence[LatestReleasedResource] = [
            model.LatestReleasedResource(
                resource_id=resource["resource_id"],
                resource_version_id=resource["resource_id"] + ",v=" + str(resource["model"]),
                id_details=data.Resource.get_details_from_resource_id(resource["resource_id"]),
                status=resource["status"],
                requires=json.loads(resource["attributes"]).get("requires", []),
            )
            for resource in records
        ]
        return dtos


class ResourcesInVersionView(DataView[VersionedResourceOrder, model.VersionedResource]):
    def __init__(
        self,
        environment: data.Environment,
        version: int,
        limit: Optional[int] = None,
        filter: Optional[Dict[str, List[str]]] = None,
        sort: str = "resource_type.desc",
        first_id: Optional[ResourceVersionIdStr] = None,
        last_id: Optional[ResourceVersionIdStr] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
    ) -> None:
        super().__init__(
            order=VersionedResourceOrder.parse_from_string(sort),
            limit=limit,
            first_id=first_id,
            last_id=last_id,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment
        self.version = version

    # Per view config
    def get_base_url(self) -> str:
        return f"/api/v2/desiredstate/{self.version}"

    @property
    def allowed_filters(self) -> Dict[str, Type[Filter]]:
        return {
            "resource_type": ContainsPartialFilter,
            "agent": ContainsPartialFilter,
            "resource_id_value": ContainsPartialFilter,
        }

    def get_base_query(self) -> SimpleQueryBuilder:
        query_builder = SimpleQueryBuilder(
            select_clause="SELECT resource_id, attributes, resource_type, agent, resource_id_value, environment",
            from_clause=f" FROM {data.Resource.table_name()}",
            filter_statements=["environment = $1", "model = $2"],
            values=[self.environment.id, self.version],
        )
        return query_builder

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[model.VersionedResource]:
        return [
            model.VersionedResource(
                resource_id=versioned_resource["resource_id"],
                resource_version_id=versioned_resource["resource_id"] + f",v={self.version}",
                id_details=data.Resource.get_details_from_resource_id(versioned_resource["resource_id"]),
                requires=json.loads(versioned_resource["attributes"]).get("requires", []),  # todo: broken
            )
            for versioned_resource in records
        ]


class CompileReportView(DataView[CompileReportOrder, CompileReport]):
    def __init__(
        self,
        environment: data.Environment,
        limit: Optional[int] = None,
        filter: Optional[Dict[str, List[str]]] = None,
        sort: str = "resource_type.desc",
        first_id: Optional[UUID] = None,
        last_id: Optional[UUID] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> None:
        super().__init__(
            order=CompileReportOrder.parse_from_string(sort),
            limit=limit,
            first_id=first_id,
            last_id=last_id,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment

    @property
    def allowed_filters(self) -> Dict[str, Type[Filter]]:
        return {
            "requested": DateRangeFilter,
            "success": BooleanEqualityFilter,
            "started": BooleanIsNotNullFilter,
            "completed": BooleanIsNotNullFilter,
        }

    def get_base_url(self) -> str:
        return "/api/v2/compilereport"

    def get_base_query(self) -> SimpleQueryBuilder:
        query_builder = SimpleQueryBuilder(
            select_clause="""SELECT id, remote_id, environment, requested,
                            started, completed, do_export, force_update,
                            metadata, environment_variables, success, version,
                            partial, removed_resource_sets, exporter_plugin,
                            notify_failed_compile, failed_compile_message""",
            from_clause=f" FROM {data.Compile.table_name()}",
            filter_statements=["environment = $1"],
            values=[self.environment.id],
        )
        return query_builder

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[model.CompileReport]:
        return [
            CompileReport(
                id=compile["id"],
                remote_id=compile["remote_id"],
                environment=compile["environment"],
                requested=compile["requested"],
                started=compile["started"],
                completed=compile["completed"],
                success=compile["success"],
                version=compile["version"],
                do_export=compile["do_export"],
                force_update=compile["force_update"],
                metadata=json.loads(compile["metadata"]) if compile["metadata"] else {},
                environment_variables=json.loads(compile["environment_variables"]) if compile["environment_variables"] else {},
                partial=compile["partial"],
                removed_resource_sets=compile["removed_resource_sets"],
                exporter_plugin=compile["exporter_plugin"],
                notify_failed_compile=compile["notify_failed_compile"],
                failed_compile_message=compile["failed_compile_message"],
            )
            for compile in records
        ]


class DesiredStateVersionView(DataView[DesiredStateVersionOrder, DesiredStateVersion]):
    def __init__(
        self,
        environment: data.Environment,
        limit: Optional[int] = None,
        filter: Optional[Dict[str, List[str]]] = None,
        sort: str = "resource_type.desc",
        start: Optional[int] = None,
        end: Optional[int] = None,
    ) -> None:
        super().__init__(
            order=DesiredStateVersionOrder.parse_from_string(sort),
            limit=limit,
            first_id=None,
            last_id=None,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment

    @property
    def allowed_filters(self) -> Dict[str, Type[Filter]]:
        return {
            "version": IntRangeFilter,
            "date": DateRangeFilter,
            "status": ContainsFilter,
        }

    def get_base_url(self) -> str:
        return "/api/v2/desiredstate"

    def get_base_query(self) -> SimpleQueryBuilder:
        subquery, subquery_values = ConfigurationModel.desired_state_versions_subquery(self.environment.id)
        query_builder = SimpleQueryBuilder(
            select_clause="SELECT *",
            from_clause=f" FROM ({subquery}) as result",
            values=subquery_values,
        )
        return query_builder

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[DesiredStateVersion]:
        return [
            DesiredStateVersion(
                version=desired_state["version"],
                date=desired_state["date"],
                total=desired_state["total"],
                labels=[DesiredStateLabel(name=desired_state["type"], message=desired_state["message"])]
                if desired_state["type"] and desired_state["message"]
                else [],
                status=desired_state["status"],
            )
            for desired_state in records
        ]


class ResourceHistoryView(DataView[ResourceHistoryOrder, ResourceHistory]):
    def __init__(
        self,
        environment: data.Environment,
        rid: ResourceIdStr,
        limit: Optional[int] = None,
        sort: str = "resource_type.desc",
        first_id: Optional[str] = None,
        last_id: Optional[str] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> None:
        super().__init__(
            order=ResourceHistoryOrder.parse_from_string(sort),
            limit=limit,
            first_id=first_id,
            last_id=last_id,
            start=start,
            end=end,
            filter={},
        )
        self.environment = environment
        self.rid = rid

    @property
    def allowed_filters(self) -> Dict[str, Type[Filter]]:
        return {}

    def get_base_url(self) -> str:
        return f"/api/v2/resource/{quote(self.rid,safe='')}/history"

    def get_base_query(self) -> SimpleQueryBuilder:
        query_builder = SimpleQueryBuilder(
            prelude="""
            /* Assign a sequence id to the rows, which is used for grouping consecutive ones with the same hash */
                WITH resourcewithsequenceids AS (
                  SELECT
                    attribute_hash,
                    model,
                    attributes,
                    date,
                    ROW_NUMBER() OVER (ORDER BY date) - ROW_NUMBER() OVER (
                      PARTITION BY attribute_hash
                      ORDER BY date
                    ) AS seqid
                  FROM resource JOIN configurationmodel cm
                    ON resource.model = cm.version AND resource.environment = cm.environment
                  WHERE resource.environment = $1 AND resource_id = $2 AND cm.released = TRUE
                )
            """,
            select_clause="SELECT attribute_hash, date, attributes, model",
            from_clause="""
            FROM (SELECT
                    attribute_hash,
                    min(date) as date,
                    min(model) as model,
                        (SELECT distinct on (attribute_hash) attributes
                            FROM resourcewithsequenceids
                            WHERE resourcewithsequenceids.attribute_hash = rs.attribute_hash
                            AND resourcewithsequenceids.seqid = rs.seqid
                            ORDER BY attribute_hash, model
                        ) as attributes
                    FROM resourcewithsequenceids rs
                    GROUP BY attribute_hash,  seqID) as sub
            """,
            values=[self.environment.id, self.rid],
        )
        return query_builder

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[ResourceHistory]:
        def get_attributes(record: Record) -> JsonType:
            attributes = json.loads(record["attributes"])
            if "version" not in attributes:
                # Due to a bug, the version field has always been present in the attributes dictionary.
                # This bug has been fixed in the database. For backwards compatibility reason we here make sure that the
                # version field is present in the attributes dictionary served out via the API.
                attributes["version"] = record["model"]
            return attributes

        return [
            ResourceHistory(
                resource_id=self.rid,
                attribute_hash=record["attribute_hash"],
                attributes=get_attributes(record),
                date=record["date"],
                requires=[Id.parse_id(rid).resource_str() for rid in json.loads(record["attributes"]).get("requires", [])],
            )
            for record in records
        ]


class ResourceLogsView(DataView[ResourceLogOrder, ResourceLog]):
    def __init__(
        self,
        environment: data.Environment,
        rid: ResourceIdStr,
        limit: Optional[int] = None,
        sort: str = "resource_type.desc",
        filter: Optional[Dict[str, List[str]]] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> None:
        super().__init__(
            order=ResourceLogOrder.parse_from_string(sort),
            limit=limit,
            first_id=None,
            last_id=None,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment
        self.rid = rid

    @property
    def allowed_filters(self) -> Dict[str, Type[Filter]]:
        return {
            "minimal_log_level": LogLevelFilter,
            "timestamp": DateRangeFilter,
            "message": ContainsPartialFilter,
            "action": ContainsFilterResourceAction,
        }

    def process_filters(self, filter: Optional[Dict[str, List[str]]]) -> Dict[str, QueryFilter]:
        # Change the api names of the filters to the names used internally in the database
        query = super().process_filters(filter)
        if query.get("minimal_log_level"):
            filter_value = query.pop("minimal_log_level")
            query["level"] = filter_value
        if query.get("message"):
            filter_value = query.pop("message")
            query["msg"] = filter_value
        return query

    def get_base_url(self) -> str:
        return f"/api/v2/resource/{parse.quote(self.rid, safe='')}/logs"

    def get_base_query(self) -> SimpleQueryBuilder:
        query_builder = SimpleQueryBuilder(
            prelude=f"""
                -- Get all resource action in the given environment for the given resource_id
                WITH actions AS (
                    SELECT  ra.*
                    FROM {Resource.table_name()} AS r INNER JOIN resourceaction_resource AS rr ON (
                                                          r.environment=rr.environment
                                                          AND r.resource_id=rr.resource_id
                                                          AND r.model=rr.resource_version
                                                      )
                                                      INNER JOIN {ResourceAction.table_name()} AS ra ON (
                                                          rr.resource_action_id=ra.action_id
                                                      )
                    WHERE r.environment=$1 AND r.resource_id=$2
                )
            """,
            select_clause="SELECT action_id, action, timestamp, unnested_message",
            from_clause="""
            FROM
                (
                    SELECT action_id,
                           action,
                           (unnested_message ->> 'timestamp')::timestamptz AS timestamp,
                           unnested_message ->> 'level' AS level,
                           unnested_message ->> 'msg' AS msg,
                           unnested_message
                    FROM actions, unnest(messages) AS unnested_message
                ) AS unnested
            """,
            values=[self.environment.id, self.rid],
        )
        return query_builder

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[ResourceLog]:
        logs = []
        for record in records:
            message = json.loads(record["unnested_message"])
            logs.append(
                ResourceLog(
                    action_id=record["action_id"],
                    action=record["action"],
                    timestamp=record["timestamp"],
                    level=message.get("level"),
                    msg=message.get("msg"),
                    args=message.get("args", []),
                    kwargs=message.get("kwargs", {}),
                )
            )
        return logs


class FactsView(DataView[FactOrder, Fact]):
    def __init__(
        self,
        environment: data.Environment,
        limit: Optional[int] = None,
        sort: str = "resource_type.desc",
        first_id: Optional[UUID] = None,
        last_id: Optional[UUID] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        filter: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        super().__init__(
            order=FactOrder.parse_from_string(sort),
            limit=limit,
            first_id=first_id,
            last_id=last_id,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment

    @property
    def allowed_filters(self) -> Dict[str, Type[Filter]]:
        return {
            "name": ContainsPartialFilter,
            "resource_id": ContainsPartialFilter,
        }

    def get_base_url(self) -> str:
        return "/api/v2/facts"

    def get_base_query(self) -> SimpleQueryBuilder:
        query_builder = SimpleQueryBuilder(
            select_clause="SELECT p.id, p.name, p.value, p.source, p.resource_id, p.updated, p.metadata, p.environment",
            from_clause=f"FROM {Parameter.table_name()} as p",
            filter_statements=["p.environment = $1 ", "p.source = 'fact'"],
            values=[self.environment.id],
        )
        return query_builder

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[Fact]:
        return [
            Fact(
                id=fact["id"],
                name=fact["name"],
                value=fact["value"],
                source=fact["source"],
                updated=fact["updated"],
                resource_id=fact["resource_id"],
                metadata=json.loads(fact["metadata"]) if fact["metadata"] else None,
                environment=fact["environment"],
            )
            for fact in records
        ]


class NotificationsView(DataView[NotificationOrder, model.Notification]):
    def __init__(
        self,
        environment: data.Environment,
        limit: Optional[int] = None,
        sort: str = "resource_type.desc",
        first_id: Optional[UUID] = None,
        last_id: Optional[UUID] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        filter: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        super().__init__(
            order=NotificationOrder.parse_from_string(sort),
            limit=limit,
            first_id=first_id,
            last_id=last_id,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment

    @property
    def allowed_filters(self) -> Dict[str, Type[Filter]]:
        return {
            "title": ContainsPartialFilter,
            "message": ContainsPartialFilter,
            "read": BooleanEqualityFilter,
            "cleared": BooleanEqualityFilter,
            "severity": ContainsFilter,
        }

    def get_base_url(self) -> str:
        return "/api/v2/notification"

    def get_base_query(self) -> SimpleQueryBuilder:
        return SimpleQueryBuilder(
            select_clause="""SELECT n.*""",
            from_clause=f" FROM {Notification.table_name()} as n",
            filter_statements=[" environment = $1 "],
            values=[self.environment.id],
        )

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[model.Notification]:
        return [
            model.Notification(
                id=notification["id"],
                title=notification["title"],
                message=notification["message"],
                severity=notification["severity"],
                created=notification["created"],
                read=notification["read"],
                cleared=notification["cleared"],
                uri=notification["uri"],
                environment=notification["environment"],
            )
            for notification in records
        ]


class ParameterView(DataView[ParameterOrder, model.Parameter]):
    def __init__(
        self,
        environment: data.Environment,
        limit: Optional[int] = None,
        sort: str = "resource_type.desc",
        first_id: Optional[UUID] = None,
        last_id: Optional[UUID] = None,
        start: Optional[Union[str, datetime]] = None,
        end: Optional[Union[str, datetime]] = None,
        filter: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        super().__init__(
            order=ParameterOrder.parse_from_string(sort),
            limit=limit,
            first_id=first_id,
            last_id=last_id,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment

    @property
    def allowed_filters(self) -> Dict[str, Type[Filter]]:
        return {
            "name": ContainsPartialFilter,
            "source": ContainsPartialFilter,
            "updated": DateRangeFilter,
        }

    def get_base_url(self) -> str:
        return "/api/v2/parameters"

    def get_base_query(self) -> SimpleQueryBuilder:
        return SimpleQueryBuilder(
            select_clause="""SELECT p.id, p.name, p.value, p.source, p.updated, p.metadata, p.environment""",
            from_clause=f"FROM {Parameter.table_name()} as p",
            filter_statements=["environment = $1", "p.source != 'fact'"],
            values=[self.environment.id],
        )

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[model.Parameter]:
        return [
            model.Parameter(
                id=parameter["id"],
                name=parameter["name"],
                value=parameter["value"],
                source=parameter["source"],
                updated=parameter["updated"],
                metadata=json.loads(parameter["metadata"]) if parameter["metadata"] else None,
                environment=parameter["environment"],
            )
            for parameter in records
        ]


class AgentView(DataView[AgentOrder, model.Agent]):
    def __init__(
        self,
        environment: data.Environment,
        limit: Optional[int] = None,
        sort: str = "resource_type.desc",
        start: Optional[Union[datetime, bool, str]] = None,
        end: Optional[Union[datetime, bool, str]] = None,
        first_id: Optional[str] = None,
        last_id: Optional[str] = None,
        filter: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        super().__init__(
            order=AgentOrder.parse_from_string(sort),
            limit=limit,
            first_id=first_id,
            last_id=last_id,
            start=start,
            end=end,
            filter=filter,
        )
        self.environment = environment

    @property
    def allowed_filters(self) -> Dict[str, Type[Filter]]:
        return {
            "name": ContainsPartialFilter,
            "process_name": ContainsPartialFilter,
            "status": ContainsFilter,
        }

    def process_filters(self, filter: Optional[Dict[str, List[str]]]) -> Dict[str, QueryFilter]:
        out_filter = super().process_filters(filter)
        # name is ambiguous, qualify
        if "name" in out_filter:
            out_filter["a.name"] = out_filter.pop("name")
        return out_filter

    def get_base_url(self) -> str:
        return "/api/v2/agents"

    def get_base_query(self) -> SimpleQueryBuilder:
        base = SimpleQueryBuilder(
            select_clause="""SELECT a.name, a.environment, last_failover, paused, unpause_on_resume,
                                            ap.hostname as process_name, ai.process as process_id,
                                            (CASE WHEN paused THEN 'paused'
                                                WHEN id_primary IS NOT NULL THEN 'up'
                                                ELSE 'down'
                                            END) as status""",
            from_clause=f" FROM {Agent.table_name()} as a LEFT JOIN public.agentinstance ai ON a.id_primary=ai.id "
            " LEFT JOIN public.agentprocess ap ON ai.process = ap.sid",
            filter_statements=[" a.environment = $1 "],
            values=[self.environment.id],
        )
        # wrap when using compound fields
        virtual_fields = {"status", "process_name", "process_id"}
        used_fields = set(self.filter.keys()).union({t[0] for t in self.order.get_order_elements(False)})
        if virtual_fields.intersection(used_fields):
            query, values = base.build()
            return SimpleQueryBuilder(
                select_clause="select *",
                from_clause=f"FROM ({query}) as a",
                values=values,
            )
        return base

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[model.Agent]:
        return [
            model.Agent(
                name=agent["name"],
                environment=agent["environment"],
                last_failover=agent["last_failover"],
                paused=agent["paused"],
                unpause_on_resume=agent["unpause_on_resume"],
                process_id=agent["process_id"],
                process_name=agent["process_name"],
                status=agent["status"],
            )
            for agent in records
        ]


class DiscoveredResourceView(DataView[DiscoveredResourceOrder, model.DiscoveredResource]):
    def __init__(
        self,
        environment: data.Environment,
        limit: Optional[int] = None,
        sort: str = "discovered_resource_id.asc",
        start: Optional[str] = None,
        end: Optional[str] = None,
    ) -> None:
        super().__init__(
            order=DiscoveredResourceOrder.parse_from_string(sort),
            limit=limit,
            first_id=None,
            last_id=None,
            start=start,
            end=end,
            filter=None,
        )
        self.environment = environment

    @property
    def allowed_filters(self) -> Dict[str, Type[Filter]]:
        """
        Return the specification of the allowed filters, see FilterValidator
        """
        return {}

    def get_base_url(self) -> str:
        return "/api/v2/discovered"

    def get_base_query(self) -> SimpleQueryBuilder:
        query_builder = SimpleQueryBuilder(
            select_clause="SELECT environment, discovered_resource_id, values",
            from_clause=f" FROM {data.DiscoveredResource.table_name()}",
            filter_statements=["environment = $1"],
            values=[self.environment.id],
        )
        return query_builder

    def construct_dtos(self, records: Sequence[Record]) -> Sequence[dict[str, str]]:
        return [
            model.DiscoveredResource(
                discovered_resource_id=res["discovered_resource_id"],
                values=json.loads(res["values"]),
            ).dict()
            for res in records
        ]



"""
    Copyright 2021 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from typing import Any, Dict, Mapping, MutableMapping, Optional, cast

from inmanta.protocol.common import JSON_CONTENT, ReturnValue, T
from inmanta.types import ArgumentTypes, ReturnTypes


class ReturnValueWithMeta(ReturnValue[T]):
    def __init__(
        self,
        status_code: int = 200,
        headers: MutableMapping[str, str] = {},
        response: Optional[T] = None,
        content_type: str = JSON_CONTENT,
        links: Optional[Dict[str, str]] = None,
        metadata: Optional[Mapping[str, ArgumentTypes]] = None,
    ) -> None:
        super().__init__(status_code, headers, response, content_type, links)
        self.metadata = metadata

    def _get_with_envelope(self, envelope_key: str) -> ReturnTypes:
        """Get the body with an envelope specified"""
        response = cast(Dict[str, Any], super()._get_with_envelope(envelope_key))
        if self.metadata:
            if response.get("metadata"):
                response["metadata"].update(self.metadata)
            else:
                response["metadata"] = self.metadata
        return response

    def _get_without_envelope(self) -> ReturnTypes:
        """Get the body without an envelope specified"""
        response = super()._get_without_envelope()
        if self.metadata:
            if isinstance(response, Dict):
                if response.get("metadata"):
                    response["metadata"].update(self.metadata)
                else:
                    response["metadata"] = self.metadata
        return response



"""
    Copyright 2022 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from asyncpg import Connection

DISABLED = False


async def update(connection: Connection) -> None:
    """
    Add the last_stable_status field in the public.resource table which
    represent the last stable state of a given resource. A stable state
    is any state different from the deploying state.
    """
    schema_updates = """
CREATE TYPE non_deploying_resource_state AS ENUM('unavailable', 'skipped', 'dry', 'deployed', 'failed', 'available',
                                                 'cancelled', 'undefined', 'skipped_for_undefined', 'processing_events');

ALTER TABLE public.resource ADD COLUMN last_non_deploying_status non_deploying_resource_state NOT NULL DEFAULT 'available';

-- Change the default value of the `last_non_deploying_status` column to the correct value.
WITH table_last_non_deploying_status AS (
    SELECT DISTINCT ON (r.environment, r.resource_version_id) r.environment, r.resource_version_id, ra.status
    FROM resource AS r INNER JOIN resourceaction AS ra
         ON r.environment=ra.environment AND ra.resource_version_ids::varchar[] @> ARRAY[r.resource_version_id]::varchar[]
    WHERE ra.status IS NOT NULL AND ra.status!='deploying'
    ORDER BY r.environment, r.resource_version_id, ra.started DESC
)
UPDATE resource AS r
SET last_non_deploying_status=s.status::text::non_deploying_resource_state
FROM table_last_non_deploying_status AS s
WHERE r.environment=s.environment AND r.resource_version_id=s.resource_version_id
    """
    async with connection.transaction():
        await connection.execute(schema_updates)



"""
    Copyright 2021 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import logging
import os
from typing import List, Optional

from inmanta.ast import Namespace
from inmanta.ast.statements import Statement
from inmanta.const import CF_CACHE_DIR, LogLevel
from inmanta.parser.pickle import ASTPickler, ASTUnpickler
from inmanta.util import get_compiler_version

LOGGER = logging.getLogger(__name__)


class CacheEnvelope:
    """Every cached file gets the exact modification time of the file it is caching, to have cheap, accurate invalidation"""

    def __init__(self, timestamp: float, statements: List[Statement]) -> None:
        self.timestamp = timestamp
        self.statements = statements


class CacheManager:
    def __init__(self) -> None:
        self.hits = 0
        self.misses = 0
        self.failures = 0

        # import loop, ....
        from inmanta.compiler.config import feature_compiler_cache

        self.cache_enabled = feature_compiler_cache
        self.root_cache_dir: Optional[str] = None

    def _get_file_name(self, namespace: Namespace, filename: str) -> str:
        """
        Returns the name for the cached file, based on the name of the original source file

        Also ensures the cache folder exists.

        :param namespace: The namespace this file is part of
        :param filename: the filename of the source file
        :return: the filename of the cached file
        """
        # Make mypy happy
        assert self.root_cache_dir is not None
        # Obtains directory where the cache file will be stored
        cache_folder = os.path.join(self.root_cache_dir, *namespace.to_path())
        # create cache folder
        os.makedirs(cache_folder, exist_ok=True)

        # get file name without extension
        filepart = os.path.basename(filename).rsplit(".", maxsplit=1)[0]
        # make filename with compiler version specific extension
        filename = f"{filepart}.{get_compiler_version().replace('.','_')}.cfc"

        # construct final path
        return os.path.join(cache_folder, filename)

    def attach_to_project(self, project_dir: str) -> None:
        if not os.path.exists(project_dir):
            raise Exception(f"Project directory {project_dir} doesn't exist")
        self.root_cache_dir = os.path.join(project_dir, CF_CACHE_DIR)

    def is_attached_to_project(self) -> bool:
        return self.root_cache_dir is not None

    def detach_from_project(self) -> None:
        self.root_cache_dir = None

    def un_cache(self, namespace: Namespace, filename: str) -> Optional[List[Statement]]:
        if not self.cache_enabled.get():
            # cache not enabled
            return None
        if not self.is_attached_to_project():
            return None
        try:
            cache_filename = self._get_file_name(namespace, filename)
            if not os.path.exists(cache_filename):
                self.misses += 1
                return None
            mtime = os.path.getmtime(filename)
            if os.path.getmtime(filename) > os.path.getmtime(cache_filename):
                self.misses += 1
                return None
            with open(cache_filename, "rb") as fh:
                result = ASTUnpickler(fh, namespace).load()
                if not isinstance(result, CacheEnvelope):
                    # old cache format
                    self.misses += 1
                    return None
                if result.timestamp != mtime:
                    # mtime is not exactly the same
                    self.misses += 1
                    return None
                self.hits += 1
                return result.statements
        except Exception:
            self.failures += 1
            LOGGER.warning(
                "Compile cache loading failure, ignoring cache entry for %s",
                filename,
                exc_info=LOGGER.isEnabledFor(LogLevel.DEBUG.to_int),
            )
            return None

    def cache(self, namespace: Namespace, filename: str, statements: List[Statement]) -> None:
        if not self.cache_enabled.get():
            # cache not enabled
            return
        if not self.is_attached_to_project():
            return
        try:
            cache_filename = self._get_file_name(namespace, filename)
            mtime = os.path.getmtime(filename)
            cache_entry = CacheEnvelope(mtime, statements)
            with open(cache_filename, "wb") as fh:
                ASTPickler(fh, protocol=4).dump(cache_entry)
        except Exception:
            LOGGER.warning(
                "Compile cache failure, failed to cache statements for %s",
                filename,
                exc_info=LOGGER.isEnabledFor(LogLevel.DEBUG.to_int),
            )

    def reset_stats(self) -> None:
        self.hits = 0
        self.misses = 0
        self.failures = 0

    def log_stats(self) -> None:
        if not self.cache_enabled.get():
            # cache not enabled
            return
        if self.hits + self.misses != 0:
            LOGGER.debug(
                "Compiler cache observed %d hits and %d misses (%d%%)",
                self.hits,
                self.misses,
                (100 * self.hits) / (self.hits + self.misses),
            )



"""
    Copyright 2020 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from io import BytesIO
from pickle import Pickler, Unpickler, UnpicklingError
from typing import Optional, Tuple

from inmanta.ast import Namespace


class ASTPickler(Pickler):
    def persistent_id(self, obj: object) -> Optional[Tuple[str, str]]:
        if isinstance(obj, Namespace):
            # Don't pickle namespaces
            return ("Namespace", obj.get_full_name())
        else:
            return None


class ASTUnpickler(Unpickler):
    def __init__(self, file: BytesIO, namespace: Namespace) -> None:
        super().__init__(file)
        self.namespace = namespace
        self.namespace_name = namespace.get_full_name()

    def persistent_load(self, pid: Tuple[str, str]) -> object:
        type_tag, key_id = pid
        if type_tag == "Namespace":
            assert self.namespace_name == key_id
            return self.namespace
        else:
            raise UnpicklingError("unsupported persistent object")



"""
    Copyright 2023 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import logging

import asyncpg

import nacl.exceptions
import nacl.pwhash
from inmanta import const, data, protocol
from inmanta.const import MIN_PASSWORD_LENGTH
from inmanta.data import AuthMethod, model
from inmanta.protocol import common, exceptions
from inmanta.server import SLICE_DATABASE, SLICE_TRANSPORT, SLICE_USER
from inmanta.server import config as server_config
from inmanta.server import protocol as server_protocol

LOGGER = logging.getLogger(__name__)


def verify_authentication_enabled() -> None:
    """raises an BadRequest exception if server authentication is not enabled"""
    if not server_config.server_enable_auth.get():
        raise exceptions.BadRequest(
            "Server authentication should be enabled. To setup the initial user use the inmanta-initial-user-setup tool."
        )


class UserService(server_protocol.ServerSlice):
    """Slice for managing users"""

    def __init__(self) -> None:
        super().__init__(SLICE_USER)

    def get_dependencies(self) -> list[str]:
        return [SLICE_DATABASE]

    def get_depended_by(self) -> list[str]:
        return [SLICE_TRANSPORT]

    @protocol.handle(protocol.methods_v2.list_users)
    async def list_users(self) -> list[model.User]:
        return [user.to_dao() for user in await data.User.get_list(order_by_column="username", order="ASC")]

    @protocol.handle(protocol.methods_v2.add_user)
    async def add_user(self, username: str, password: str) -> model.User:
        verify_authentication_enabled()
        if not username:
            raise exceptions.BadRequest("the username cannot be an empty string")
        if not password or len(password) < MIN_PASSWORD_LENGTH:
            raise exceptions.BadRequest("the password should be at least 8 characters long")

        # hash the password
        pw_hash = nacl.pwhash.str(password.encode())

        # insert the user
        try:
            user = data.User(
                username=username,
                password_hash=pw_hash.decode(),
                auth_method=AuthMethod.database,
            )
            await user.insert()
        except asyncpg.UniqueViolationError:
            raise exceptions.Conflict(f"A user with name {username} already exists.")
        return user.to_dao()

    @protocol.handle(protocol.methods_v2.delete_user)
    async def delete_user(self, username: str) -> None:
        verify_authentication_enabled()
        user = await data.User.get_one(username=username)
        if user is None:
            raise exceptions.NotFound(f"User with name {username} does not exist.")

        await user.delete()

    @protocol.handle(protocol.methods_v2.set_password)
    async def set_password(self, username: str, password: str) -> None:
        verify_authentication_enabled()
        if not password or len(password) < MIN_PASSWORD_LENGTH:
            raise exceptions.BadRequest("the password should be at least 8 characters long")
        # check if the user already exists
        user = await data.User.get_one(username=username)
        if not user:
            raise exceptions.NotFound(f"User with name {username} does not exist.")

        # hash the password
        pw_hash = nacl.pwhash.str(password.encode())

        # insert the user
        await user.update_fields(password_hash=pw_hash.decode())

    @protocol.handle(protocol.methods_v2.login)
    async def login(self, username: str, password: str) -> common.ReturnValue[model.LoginReturn]:
        verify_authentication_enabled()
        # check if the user exists
        user = await data.User.get_one(username=username)
        if not user or not user.password_hash:
            raise exceptions.UnauthorizedException()
        try:
            nacl.pwhash.verify(user.password_hash.encode(), password.encode())
        except nacl.exceptions.InvalidkeyError:
            raise exceptions.UnauthorizedException()

        token = common.encode_token([str(const.ClientType.api.value)], expire=None)
        return common.ReturnValue(
            status_code=200,
            headers={"Authentication": f"Bearer {token}"},
            response=model.LoginReturn(
                user=user.to_dao(),
                token=token,
            ),
        )



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import os
import re
from itertools import groupby

import pytest

import inmanta.compiler as compiler
from inmanta.ast import AttributeException, RuntimeException


@pytest.fixture
def setup_project_for(snippetcompiler):
    def do_setup(name: str, main_file: str = "main.cf"):
        path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", name)
        return snippetcompiler.setup_for_existing_project(path, main_file=main_file)

    return do_setup


def test_compile_test_1(setup_project_for):
    setup_project_for("compile_test_1")
    (types, _) = compiler.do_compile()
    instances = types["__config__::Host"].get_all_instances()
    assert len(instances) == 1
    i = instances[0]
    assert i.get_attribute("name").get_value() == "test1"
    assert i.get_attribute("os").get_value().get_attribute("name").get_value() == "linux"


def test_compile_test_2(setup_project_for):
    setup_project_for("compile_test_2")
    (types, _) = compiler.do_compile()
    instances = types["__config__::ManagedDevice"].get_all_instances()
    assert sorted([i.get_attribute("name").get_value() for i in instances]) == [1, 2, 3, 4, 5]


def test_compile_test_index_collission(setup_project_for):
    setup_project_for("compile_test_index_collission")
    with pytest.raises(RuntimeException):
        compiler.do_compile()


def test_compile_test_index(setup_project_for):
    setup_project_for("compile_test_index")
    (_, scopes) = compiler.do_compile()
    variables = {k: x.get_value() for k, x in scopes.get_child("__config__").scope.slots.items()}

    p = re.compile(r"(f\d+h\d+)(a\d+)?")

    items = [
        (m.groups()[0], m.groups()[1], v) for m, v in [(re.search(p, k), v) for k, v in variables.items()] if m is not None
    ]
    groups = groupby(sorted(items, key=lambda x: x[0]), lambda x: x[0])
    firsts = []
    for k, v in groups:
        v = list(v)
        first = v[0]
        firsts.append(first)
        for other in v[1:]:
            assert first[2] == other[2]

    for i in range(len(firsts)):
        for j in range(len(firsts)):
            if not i == j:
                assert firsts[i][2] != firsts[j][2], "Variable %s%s should not be equal to %s%s" % (
                    firsts[i][0],
                    firsts[i][1],
                    firsts[j][0],
                    firsts[j][1],
                )


def test_compile_test_double_assign(setup_project_for):
    setup_project_for("compile_test_double_assign")
    with pytest.raises(AttributeException):
        compiler.do_compile()


def test_compile_138(setup_project_for):
    setup_project_for("compile_138")
    (types, _) = compiler.do_compile()
    assert (
        types["std::Host"].get_all_instances()[0].get_attribute("agent").get_value().get_attribute("names").get_value()
        is not None
    )


def test_compile_plugin_typing(setup_project_for):
    setup_project_for("compile_plugin_typing")
    (_, scopes) = compiler.do_compile()
    root = scopes.get_child("__config__")

    def verify(name):
        c1a1 = root.lookup(name).get_value()
        name = sorted([item.get_attribute("name").get_value() for item in c1a1])
        assert name == ["t1", "t2", "t3"]

    verify("c1a1")
    verify("c1a2")

    s1 = root.lookup("s1").get_value()
    s2 = root.lookup("s2").get_value()

    assert s2[0] == s1
    assert isinstance(s2, list)
    assert isinstance(s2[0], str)


def test_compile_plugin_typing_invalid(setup_project_for):
    project = setup_project_for("compile_plugin_typing", "invalid.cf")
    with pytest.raises(RuntimeException) as e:
        compiler.do_compile()
    text = e.value.format_trace(indent="  ")
    print(text)
    assert (
        text
        == """Exception in plugin test::badtype (reported in test::badtype(c1.items) ({dir}/invalid.cf:16))
caused by:
  Invalid type for value 'a', should be type test::Item (reported in test::badtype(c1.items) ({dir}/invalid.cf:16))""".format(
            dir=project.project_path
        )
    )
